#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2010 - 2021 Tatsuhiro Tsujikawa
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
# nala is based upon apt-metalink https://github.com/tatsuhiro-t/apt-metalink
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Functions for the Nala Install command."""
from __future__ import annotations

import contextlib
import fcntl
import hashlib
import os
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import Iterable, List, Sequence, cast

import apt_pkg
from apt.cache import FetchFailedException, LockFailedException
from apt.package import BaseDependency, Dependency, Package, Version
from apt_pkg import DepCache, Error as AptError, get_architectures
from httpx import HTTPError, head

from nala import _, color, color_version
from nala.cache import Cache
from nala.constants import (
	ARCHIVE_DIR,
	DPKG_LOG,
	ERROR_PREFIX,
	NALA_DIR,
	NALA_TERM_LOG,
	NOTICE_PREFIX,
	REBOOT_PKGS,
	REBOOT_REQUIRED,
	WARNING_PREFIX,
	CurrentState,
)
from nala.debfile import NalaBaseDep, NalaDebPackage, NalaDep
from nala.downloader import (
	URL,
	Downloader,
	URLSet,
	download,
	download_pkgs,
	print_error,
)
from nala.dpkg import DpkgLive, InstallProgress, OpProgress, UpdateProgress, notice
from nala.error import (
	BrokenError,
	ExitCode,
	apt_error,
	essential_error,
	local_deb_error,
	print_dpkg_errors,
)
from nala.history import write_history
from nala.options import arguments
from nala.rich import ELLIPSIS, Text, dpkg_progress, from_ansi
from nala.summary import print_update_summary
from nala.utils import (
	DelayedKeyboardInterrupt,
	NalaPackage,
	PackageHandler,
	ask,
	dprint,
	eprint,
	get_date,
	get_pkg_name,
	get_pkg_version,
	pkg_installed,
	term,
	unauth_ask,
	vprint,
)

# pylint: disable=too-many-lines


def auto_remover(cache: Cache, nala_pkgs: PackageHandler, config: bool = False) -> None:
	"""Handle auto removal of packages."""
	dprint("Starting Auto Remover")
	if not arguments.auto_remove and arguments.command not in (
		"autoremove",
		"autopurge",
	):
		dprint("Packages will not be autoremoved")
		nala_pkgs.not_needed = [
			NalaPackage(pkg.name, pkg.installed.version, pkg.installed.installed_size)
			for pkg in cache
			if pkg.installed and not pkg.marked_delete and pkg.is_auto_removable
		]
		return

	dprint("Auto-Removing")
	with cache.actiongroup():  # type: ignore[attr-defined]
		# Recurse 10 levels if we're installing .debs to make sure that all depends are safe
		deps = recurse_deps(nala_pkgs.local_debs, levels=10, installed=False)
		for pkg in cache:
			if pkg.is_installed and not pkg.marked_delete and pkg.is_auto_removable:
				if pkg in deps:
					dprint(f"Dependency: ['{pkg.name}'] Protected from removal")
					continue
				# We don't have to autofix while autoremoving
				pkg.mark_delete(auto_fix=False, purge=arguments.is_purge())
				nala_pkgs.autoremoved.add(pkg.name)

			elif config and not pkg.is_installed and pkg.has_config_files:
				vprint(
					_("Purging configuration files for {package}").format(
						package=color(pkg.name, "RED")
					)
				)
				pkg.mark_delete(auto_fix=False, purge=arguments.is_purge())
				nala_pkgs.autoremoved.add(pkg.name)

	dprint(f"Pkgs marked by autoremove: {nala_pkgs.autoremoved}")


def recurse_deps(
	pkgs: Iterable[NalaDebPackage] | Iterable[Package],
	levels: int = 1,
	installed: bool = False,
) -> set[Package]:
	"""Return the installed dependency packages.

	This function can be used recursively.

	Example for recursing deps 2 levels, returning only those installed::

	        deps = installed_deps(list[Package], 2, installed=True)

	Args::

	        pkgs (list[NalaDebPackage] | list[Package]):  list of package objects.
	        recurse (int): How many levels to traverse dependencies. Default is 1.
	        installed (bool): Whether to grab dependencies that are installed or all. Default False.
	"""
	if not pkgs:
		# Return an empty list if we didn't get packages passed.
		return set()
	total_deps: set[Package] = set()
	for _ in range(levels):
		dep_list: set[Package] = set()
		for dpkg in pkgs:
			dependencies = get_dep_type(dpkg, installed)
			for deps in dependencies:
				# deps len greater than 1 are or_deps
				if len(deps) > 1:
					for ndep in deps:
						dep_list |= get_dep_pkgs(ndep, installed)
					continue
				dep_list |= get_dep_pkgs(deps[0], installed)
		total_deps |= dep_list
		pkgs = dep_list
	dprint(
		f"Recurse Levels: {levels}, Recursive List Size: {len(total_deps)}, "
		f"Recurse Type: {'Installed' if installed else 'All Packages'}"
	)
	return total_deps


def get_dep_pkgs(
	ndep: NalaBaseDep | BaseDependency, installed: bool = False
) -> set[Package]:
	"""Return the packages of the specified dep that we are to use."""
	target_versions = (
		ndep.installed_target_versions if installed else ndep.target_versions
	)
	return {version.package for version in target_versions}


def get_dep_type(
	dpkg: NalaDebPackage | Package, installed: bool = False
) -> list[Dependency] | list[NalaDep]:
	"""Return the installed or candidate dependencies."""
	if isinstance(dpkg, Package):
		# We know it is installed as we check outside this function
		if installed and dpkg.installed:
			return dpkg.installed.dependencies
		if not installed and dpkg.candidate:
			return dpkg.candidate.dependencies
		return cast(List[Dependency], [])
	return dpkg.dependencies


def fix_excluded(protected: set[Package], is_upgrade: Iterable[Package]) -> list[str]:
	"""Find and optionally fix packages that need protecting."""
	eprint(
		_("{notice} Selected packages cannot be excluded from upgrade safely.").format(
			notice=NOTICE_PREFIX
		)
	)
	new_pkg = set()
	old_pkg = set()
	for pkg in protected:
		old_pkg.add(pkg.name)
		if not pkg.candidate:
			continue
		for deps in pkg.candidate.dependencies:
			for base_dep in deps:
				if base_dep.target_versions:
					dep_pkg = base_dep.target_versions[0].package
					if (
						dep_pkg in is_upgrade
						and dep_pkg.marked_install
						or dep_pkg.marked_upgrade
					):
						new_pkg.add(dep_pkg.name)
	if not new_pkg:
		eprint(
			_(
				"{error} Unable to calculate how to protect the selected packages"
			).format(error=ERROR_PREFIX)
		)
		sys.exit(_("{error} You have held broken packages").format(error=ERROR_PREFIX))
	eprint(
		_("{notice} The following packages need to be protected as well:").format(
			notice=NOTICE_PREFIX
		)
	)
	eprint(
		f"  {' '.join(color(name, 'YELLOW') for name in sorted(new_pkg) if name not in old_pkg)}\n"
	)
	return sorted(new_pkg | old_pkg)


def set_comp(current_version: Version, cand: Version) -> str:
	"""Set the compare string."""
	if current_version < cand:
		return f"< {cand.version} "
	if current_version > cand:
		return f"> {cand.version} "
	# It must be equals
	return f"= {cand.version} "


def set_multi_arch(version: Version, hook_ver: int) -> str:
	"""Set multi arch if Version 3."""
	if hook_ver >= 3:
		return f"{version.architecture} {version._cand.multi_arch} "
	return ""


def get_now_version(pkg: Package) -> Version | None:
	"""Get the now Version or None."""
	for ver in pkg.versions:
		for origin in ver.origins:
			if origin.archive == "now":
				return ver
	return None


def pkg_info(pkg: Package, version: int) -> str:
	"""Set package info for version 2+."""
	string = ""
	# Set currently installed version.
	if not (current_version := pkg.installed) and pkg.marked_delete:
		current_version = get_now_version(pkg)

	string += f"{pkg.name} "

	if current_version:
		file = ARCHIVE_DIR / get_pkg_name(current_version)
		string += (
			f"{current_version.version} {set_multi_arch(current_version, version)}"
		)
	else:
		string += "- " if version <= 2 else "- - none "

	if cand := pkg.candidate:
		file = ARCHIVE_DIR / get_pkg_name(cand)
		if current_version:
			string += (
				f"{set_comp(current_version, cand)} {set_multi_arch(cand, version)}"
			)
		else:
			string += f"< {cand.version} {set_multi_arch(cand, version)}"
	else:
		string += "> - " if version <= 2 else "> - - none "

	if pkg.marked_install or pkg.marked_upgrade:
		# It is not be able to reach this code without being `file ` assigned.
		# pylint: disable=possibly-used-before-assignment
		string += f"{file}\n" if file else "**ERROR**\n"  #
	elif pkg.marked_delete:
		string += "**REMOVE**\n"
	elif pkg.has_config_files:
		string += "**CONFIGURE**\n"
	else:
		string += f"{pkg.marked_upgrade}\n"
	return string


def write_config_info(w: TextIOWrapper, version: int) -> None:
	"""Seend the version and config info to the hook."""
	# Send the Hook the Version of apt hook it will use.
	# This version of APT supports only v3, so don't sent higher versions
	if version <= 3:
		w.write(f"VERSION {version}\n")
	else:
		w.write("VERSION 3\n")
	w.flush()

	# Write out configuration to the hook
	for line in arguments.config.apt.dump().splitlines():
		key, value = line.split(maxsplit=1)
		# Strip config formatters and remove any empty values
		if raw_value := value.strip('";'):
			key = apt_pkg.quote_string(key, '="\n')  # type: ignore[attr-defined]
			value = apt_pkg.quote_string(raw_value, "\n")  # type: ignore[attr-defined]

			w.write(f"{key}={value}\n")
			w.flush()
	w.write("\n")
	w.flush()


def apt_hook_with_pkgs(cache: Cache) -> None:
	"""Run apt hooks with packages."""
	apt_hooks = apt_pkg.config.value_list("DPkg::Pre-Install-Pkgs")
	# Remove the hooks so that apt doesn't also run them.
	apt_pkg.config.clear("DPkg::Pre-Install-Pkgs")

	for hook in apt_hooks:
		if not hook:
			continue

		args = hook.split()

		version = arguments.config.apt.find_i(  # type: ignore[attr-defined]
			f"DPkg::Tools::Options::{args[0]}::VERSION", 1
		)
		info_fd = arguments.config.apt.find_i(  # type: ignore[attr-defined]
			f"DPkg::Tools::Options::{args[0]}::InfoFD", 0
		)

		# Setup package information
		pkgs: list[str] = []

		for pkg in cache.get_changes():
			if version <= 1:
				# Only deal with packages marked install or upgraded
				if not (pkg.marked_install or pkg.marked_upgrade):
					continue

				if not (cand := pkg.candidate):
					continue

				# Make sure the file exists
				if not (file := ARCHIVE_DIR / get_pkg_name(cand)):
					continue
				pkgs.append(f"{file}\n")
				continue

			pkgs.append(pkg_info(pkg, version))

		# Start setting up pipes
		(statusfd, writefd) = os.pipe()

		if pid := os.fork() == 0:
			os.set_inheritable(statusfd, True)
			os.dup2(statusfd, info_fd)

			fcntl.fcntl(statusfd, fcntl.F_SETFL, os.O_NONBLOCK)
			os.environ["APT_HOOK_INFO_FD"] = f"{info_fd}"
			args = ["/bin/sh", "-c", hook]
			os._exit(os.execv(args[0], args))
		else:
			w = os.fdopen(writefd, "w")

			# Don't write config data for Version 1
			if version >= 2:
				write_config_info(w, version)

			for pkg in pkgs:  # type: ignore[assignment]
				w.write(pkg)  # type: ignore[arg-type]
				w.flush()
			w.close()

			# We need to exit if a Hook ends up failing.
			if exit_code := os.WEXITSTATUS(
				# Wait for the pid to finish and get it's exit code.
				os.waitpid(pid, 0)[1]
			):
				sys.exit(exit_code)


def run_scripts(hooks: list[str]) -> None:
	"""Run system scripts."""
	for hook in hooks:
		if exit_code := os.WEXITSTATUS(os.system(hook)):
			sys.exit(exit_code)


def commit_pkgs(cache: Cache, nala_pkgs: PackageHandler) -> None:
	"""Commit the package changes to the cache."""
	dprint("Commit Pkgs")
	task = dpkg_progress.add_task("", total=nala_pkgs.dpkg_progress_total())

	pre_invoke = arguments.config.apt.value_list("DPkg::Pre-Invoke")
	arguments.config.apt.clear("DPkg::Pre-Invoke")

	post_invoke = arguments.config.apt.value_list("DPkg::Post-Invoke")
	arguments.config.apt.clear("DPkg::Post-Invoke")

	run_scripts(pre_invoke)
	apt_hook_with_pkgs(cache)

	with DpkgLive(install=True) as live:
		with open(DPKG_LOG, "w", encoding="utf-8") as dpkg_log:
			with open(NALA_TERM_LOG, "a", encoding="utf-8") as term_log:
				term_log.write(
					_("Log Started: [{date}]").format(date=get_date()) + "\n"
				)
				if arguments.raw_dpkg:
					live.stop()
				config_purge = tuple(
					pkg.name
					for pkg in nala_pkgs.autoremove_config + nala_pkgs.delete_config
				)
				install = InstallProgress(dpkg_log, term_log, live, task, config_purge)
				update = UpdateProgress(live)
				cache.commit_pkgs(install, update)
				if nala_pkgs.local_debs:
					cache.commit_pkgs(install, update, nala_pkgs.local_debs)
				term_log.write(
					_("Log Ended: [{date}]").format(date=get_date()) + "\n\n"
				)
		# If we made it this far just set the total to 100%.
		dpkg_progress.reset(task)
		dpkg_progress.advance(task, advance=nala_pkgs.dpkg_progress_total())
		live.scroll_bar(rerender=True)
	run_scripts(post_invoke)


def get_changes(cache: Cache, nala_pkgs: PackageHandler, operation: str) -> None:
	"""Get packages requiring changes and process them."""
	cache.purge_removed()
	pkgs = sorted(cache.get_changes(), key=sort_pkg_name)
	if not NALA_DIR.exists():
		NALA_DIR.mkdir()

	if operation not in ("upgrade", "remove"):
		if not arguments.install_recommends:
			get_extra_pkgs(
				"Recommends", pkgs + nala_pkgs.local_debs, nala_pkgs.recommend_pkgs  # type: ignore[operator]
			)
		if not arguments.install_suggests:
			get_extra_pkgs(
				"Suggests", pkgs + nala_pkgs.local_debs, nala_pkgs.suggest_pkgs  # type: ignore[operator]
			)

	check_work(pkgs, nala_pkgs, operation)

	if pkgs or nala_pkgs.local_debs or nala_pkgs.configure_pkgs:
		check_essential(pkgs)
		sort_pkg_changes(pkgs, nala_pkgs)
		print_update_summary(nala_pkgs, cache)
		check_term_ask()

	# Enable verbose and raw_dpkg if we're piped.
	if not term.can_format():
		arguments.verbose = True
		arguments.raw_dpkg = True
	# If we're in Raw_Dpkg we can restore the locale as Nala doesn't handle the output
	if arguments.raw_dpkg:
		term.restore_locale()

	download_pkgs(pkgs)
	start_dpkg(cache, nala_pkgs)
	write_history(cache, nala_pkgs, operation)


def start_dpkg(cache: Cache, nala_pkgs: PackageHandler) -> None:
	"""Start dpkg."""
	try:
		# Set Use-Pty to False. This makes Sigwinch signals accepted by dpkg.
		apt_pkg.config.set("Dpkg::Use-Pty", "0")
		commit_pkgs(cache, nala_pkgs)
	# Catch system error because if dpkg fails it'll throw this
	except (apt_pkg.Error, SystemError) as error:
		apt_error(error)
		sys.exit(1)
	except FetchFailedException as error:
		# We have already printed this error likely. but just in case
		# We write it to the dpkg_log so at least we'll know about it.
		with open(DPKG_LOG, "a", encoding="utf-8") as file:
			file.write("FetchedFailedException:\n")
			file.write(f"{error}")
		eprint(_("{error} Fetching packages has failed!").format(error=ERROR_PREFIX))
		sys.exit(1)
	except KeyboardInterrupt:
		eprint(_("Exiting due to SIGINT"))
		sys.exit(ExitCode.SIGINT)
	finally:
		term.restore_mode()
		# If dpkg quits for any reason we lose the cursor
		if term.can_format():
			term.write(term.SHOW_CURSOR)

		print_dpkg_errors()
		print_notices(notice)
		if need_reboot():
			print(_("{notice} A reboot is required.").format(notice=NOTICE_PREFIX))
	print(color(_("Finished Successfully"), "GREEN"))


def install_local(nala_pkgs: PackageHandler, cache: Cache) -> None:
	"""Mark the depends for local debs to be installed.

	Dependencies that are marked will be marked auto installed.

	Returns local_names to be printed in the transaction summary.
	"""
	failed: list[NalaDebPackage] = []
	for pkg in nala_pkgs.local_debs[:]:
		if not pkg.check(allow_downgrade=True):
			failed.append(pkg)
			nala_pkgs.local_debs.remove(pkg)
			continue

		if not check_local_version(pkg, nala_pkgs):
			nala_pkgs.install_pkgs.append(
				NalaPackage(
					pkg.pkgname,
					pkg._sections["Version"],
					pkg.installed_size(),
				)
			)

		depends = pkg.dependencies
		extra_pkgs = (
			(arguments.install_recommends, pkg.get_dependencies("Recommends")),
			(arguments.install_suggests, pkg.get_dependencies("Suggests")),
		)

		# Check to see if we need to install any recommends or suggests.
		for check, extra_deps in extra_pkgs:
			if not check:
				continue

			for dep in extra_deps:
				# If one of the deps are already installed we don't need to do anything
				if dep.installed_target_versions:
					continue

				candidate = None
				# Check providers first to make sure we get the best package.
				if providers := cache.get_providing_packages(
					dep[0].name, include_nonvirtual=True
				):
					candidate = providers[0]

				# If there are no providers just try the cache
				elif dep[0].name in cache:
					candidate = cache[dep[0].name]

				if candidate and not candidate.installed:
					candidate.mark_install(from_user=False)
					depends.append(dep)

		satisfy_notice(pkg, depends)

	if failed:
		BrokenError(cache, failed).broken_install()


def satisfy_notice(pkg: NalaDebPackage, depends: list[NalaDep]) -> None:
	"""Print a notice of how to satisfy the packages dependencies."""
	fixer: list[str] = []

	for dep in depends:
		fixer.extend(
			color(ppkg.name, "GREEN")
			for base_dep in dep
			if (target := list(base_dep.target_versions))
			and (ppkg := target[0].package).marked_install
		)

	if fixer:
		print(
			_("{notice} The following will be installed to satisfy {package}:").format(
				notice=NOTICE_PREFIX, package=color(pkg.name, "GREEN")
			)
		)
		print(f"  {', '.join(fixer)}")


def check_local_version(pkg: NalaDebPackage, nala_pkgs: PackageHandler) -> bool:
	"""Check if the version installed is better than the .deb.

	Return True if we've added to a package list, False if not.

	VERSION_NONE = 0
	VERSION_OUTDATED = 1
	VERSION_SAME = 2
	VERSION_NEWER = 3
	"""
	if pkg_compare := pkg.compare_to_version_in_cache():
		cache_pkg = pkg._cache[pkg.pkgname]
		dprint(f"Comparing cache versions of: {pkg.pkgname}")
		# Example filename ../scar-archive/nala_0.5.0_amd64.deb
		assert pkg.filename
		dprint(f"Filename: {(filename := pkg.filename.split('/')[-1])}")
		dprint(f"Cache pkg: {cache_pkg.fullname}")
		if pkg_compare == pkg.VERSION_SAME and cache_pkg.is_installed:
			dprint(f"{filename} is the same version as the installed pkg")
			nala_pkgs.reinstall_pkgs.append(
				NalaPackage(
					pkg.pkgname,
					pkg._sections["Version"],
					pkg.installed_size(),
					pkg_installed(cache_pkg).version,
				)
			)

			if (
				pkg.compare_to_version_in_cache(use_installed=False)
				== pkg.VERSION_OUTDATED
			):
				if not cache_pkg.candidate:
					return True
				color_name = color(cache_pkg.name, "GREEN")
				print(
					_(
						"{notice} Newer version {package} {version} exists in the cache.\n"
						"You should consider using `{command}`"
					).format(
						notice=NOTICE_PREFIX,
						package=color_name,
						version=color_version(cache_pkg.candidate.version),
						command=f"{color('nala install')} {color_name}",
					)
				)
			return True

		if pkg_compare == pkg.VERSION_OUTDATED:
			dprint(f"{pkg.filename} is an older version than the installed pkg")
			nala_pkgs.downgrade_pkgs.append(
				NalaPackage(
					pkg.pkgname,
					pkg._sections["Version"],
					pkg.installed_size(),
					pkg_installed(cache_pkg).version,
				)
			)
			return True
		if pkg_compare == pkg.VERSION_NEWER and cache_pkg.is_installed:
			dprint(f"{pkg.filename} is a newer version than the installed pkg")
			nala_pkgs.upgrade_pkgs.append(
				NalaPackage(
					pkg.pkgname,
					pkg._sections["Version"],
					pkg.installed_size(),
					pkg_installed(cache_pkg).version,
				)
			)
			return True
	return False


def prioritize_local(
	deb_pkg: NalaDebPackage, cache_name: str, pkg_names: list[str]
) -> None:
	"""Print a notice of prioritization and remove the pkg name from list."""
	assert deb_pkg.filename
	print(
		_("{notice} {deb} has taken priority over {package} from the cache.").format(
			notice=NOTICE_PREFIX,
			deb=color(deb_pkg.filename.split("/")[-1], "GREEN"),
			package=color(cache_name, "YELLOW"),
		)
	)
	pkg_names.remove(cache_name)


def get_url_size(url: str) -> int:
	"""Get the URL Header and check for content length."""
	# We must get the headers so we know what the filesize is.
	response = head(url, follow_redirects=True)
	response.raise_for_status()
	dprint(response.headers)

	try:
		return int(response.headers["content-length"])
	except KeyError:
		sys.exit(
			_(
				"{error} No content length in response from {url}\n"
				"  Ensure the URL points to a Debian Package"
			).format(error=ERROR_PREFIX, url=url)
		)


def split_url(url_string: str, cache: Cache) -> URLSet:
	"""Split the URL and try to determine the hash."""
	dprint(url_split := url_string.split(":"))

	# http
	proto = url_split[0]
	# //deb.debian.org/debian/pool/main/n/neofetch/neofetch_7.1.0-2_all.deb
	body = url_split[1]

	# neofetch_7.1.0-2_all.deb
	dprint(f"Filename: {(filename := body.split('/').pop())}")

	# Initialize a URL
	url = URL(
		f"{proto}:{body}",
		get_url_size(f"{proto}:{body}"),
		ARCHIVE_DIR / filename,
		proto,
	)

	pkg_attrs = filename.split("_")
	# ["neofetch", "7.1.0-2", "all"]
	dprint(f"Package Name: {(pkg_attrs[0])}\n")

	try:
		hash_or_type = url_split[2]
	# IndexError Occurs because they did not specify a hash
	except IndexError:
		eprint(
			_("{notice} {filename} can't be hashsum verified.").format(
				notice=NOTICE_PREFIX, filename=filename
			)
		)

		if not unauth_ask(_("Do you want to continue?")):
			sys.exit(_("Abort."))

		url.no_hash = True
		return URLSet([url])

	# sha512 d500faf8b2b9ee3a8fbc6a18f966076ed432894cd4d17b42514ffffac9ee81ce
	# 945610554a11df24ded152569b77693c57c7967dd71f644af3066bf79a923bfe
	#
	# sha256 a694f44fa05fff6d00365bf23217d978841b9e7c8d7f48e80864df08cebef1a8
	# md5 b9ef863f210d170d282991ad1e0676eb
	# sha1 d1f34ed00dea59f886b9b99919dfcbbf90d69e15

	# Length of the hex digests
	len_map = {
		128: "sha512",
		64: "sha256",
		32: "md5sum",
		40: "sha1",
	}

	# Clear the hash_type
	url.hash_type = ""
	# Attempt to autodetect the hash type based on the len of the hash
	if hash_type := len_map.get(len(hash_or_type)):
		url.hash_type = hash_type
		url.hash = hash_or_type
		vprint(f"Automatically Selecting {url.hash_type}: {url.hash}")

	# If it doesn't match then it must be specified
	else:
		url.hash_type = hash_or_type

		try:
			url.hash = url_split[3]
			# This is just testing to ensure it's supported
			hashlib.new(url.hash_type)

			# If the Type is known we can check the length of the hash to ensure that it's proper
			if (
				url.hash_type in len_map.values()
				and url.hash_type != len_map[len(url.hash)]
			):
				sys.exit(
					_("{error} Hash does not match the '{hash_type}' Length").format(
						error=ERROR_PREFIX, hash_type=color(url.hash_type, "YELLOW")
					)
				)

		except IndexError:
			sys.exit(
				_("{error} Hash Type '{hash_type}' specified with no hash").format(
					error=ERROR_PREFIX, hash_type=color(url.hash_type, "YELLOW")
				)
			)

		except ValueError:
			sys.exit(
				_("{error} Hash Type '{hash_type}' is unsupported").format(
					error=ERROR_PREFIX, hash_type=color(url.hash_type, "YELLOW")
				)
			)

	# Everything hashed out, lets check for any extra URI's we can add
	dprint(url)
	url_set = URLSet([url])

	# Check to see if our package is in the cache
	if pkg_attrs[0] in cache:
		dprint("Package found in the cache")
		pkg = cache[pkg_attrs[0]]
		for ver in pkg.versions:
			hash_list = ver._records.hashes
			with contextlib.suppress(KeyError):
				if url.hash == hash_list.find(url.hash_type).hashvalue:
					dprint("Package Hash Found in the cache. Adding URIs")
					url_set.append(URL.from_version(ver))

	# You can check the versions with this. I don't know if it's useful yet
	# pylint: disable=line-too-long
	# if (cmp := apt_pkg.version_compare(version, ver.version)) > 0:
	# 	eprint("A nice little upgrade!")
	# elif cmp < 0:
	# 	eprint(_("{notice} Woah are you planning on downgrading this package?").format(notice=NOTICE_PREFIX))
	# else:
	# 	print("The Versions are identical")
	return url_set


def split_local(
	pkg_names: list[str], cache: Cache, local_debs: list[NalaDebPackage]
) -> list[str]:
	"""Split pkg_names into either Local debs, regular install or they don't exist."""
	not_exist: list[str] = []
	download_debs = []
	if urls := [name for name in pkg_names if name.startswith(("http://", "https://"))]:
		print(f"Checking Urls{ELLIPSIS}")
		for url in urls:
			try:
				vprint(f"Verifying {url}")
				url_set = split_url(url, cache)
			except HTTPError as error:
				print_error(error)
				sys.exit(1)

			download_debs.append(url_set)
			pkg_names.remove(url)
			pkg_names.append(f"{url_set.path()}")

	# .deb packages have to be downloaded before anything else in order to determine dependencies
	if download_debs:
		download(Downloader(download_debs))

	for name in pkg_names[:]:
		if ".deb" in name or "/" in name:
			path = Path(name)
			if not path.exists():
				not_exist.append(name)
				pkg_names.remove(name)
				continue
			try:
				local_debs.append(deb_pkg := NalaDebPackage(name, cache))
			except AptError as error:
				local_deb_error(error, name)

			if deb_pkg.pkgname in pkg_names and deb_pkg.pkgname in cache:
				prioritize_local(deb_pkg, deb_pkg.pkgname, pkg_names)
			for arch in get_architectures():
				if (
					arch_pkg := f"{deb_pkg.pkgname}:{arch}"
				) in pkg_names and arch_pkg in cache:
					prioritize_local(deb_pkg, arch_pkg, pkg_names)

			pkg_names.remove(name)
			continue
	return not_exist


def package_manager(pkg_names: list[str], cache: Cache, remove: bool = False) -> bool:
	"""Manage installation or removal of packages."""
	with cache.actiongroup():  # type: ignore[attr-defined]
		fixer = apt_pkg.ProblemResolver(cache._depcache)
		for pkg_name in pkg_names:
			if pkg_name in cache:
				pkg = cache[pkg_name]
				try:
					if remove:
						if pkg.installed or (
							pkg.has_config_files and arguments.is_purge()
						):
							pkg.mark_delete(
								auto_fix=arguments.fix_broken,
								purge=arguments.is_purge(),
							)
							dprint(f"Marked Remove: {pkg.name}")
						continue
					if not pkg.installed or pkg.marked_downgrade:
						# Auto_inst is false as we need to do this part later
						# after all packages have been marked.
						pkg.mark_install(auto_inst=False, auto_fix=False)
						fixer.clear(pkg._pkg)
						fixer.protect(pkg._pkg)
						dprint(f"Marked Install: {pkg.name}")
					elif pkg.is_upgradable:
						pkg.mark_upgrade()
						dprint(f"Marked upgrade: {pkg.name}")
				except AptError as error:
					if (
						"broken packages" not in f"{error}"
						and "held packages" not in f"{error}"
					):
						raise error from error
					return False
	# When installing packages we need to iterate them again and mark them differently
	# Apt does not do this for removing packages.
	# https://github.com/volitank/nala/issues/27
	if not remove:
		for pkg_name in pkg_names:
			if pkg_name in cache:
				pkg = cache[pkg_name]
				pkg.mark_install(auto_fix=arguments.fix_broken)
	return True


def set_candidate_versions(
	pkg_names: list[str], cache: Cache
) -> tuple[list[str], bool]:
	"""Set the version to be installed."""
	not_found: list[str] = []
	failed = False
	for name in pkg_names[:]:
		if "=" not in name:
			continue
		pkg_name, version = name.split("=")

		if pkg_name not in cache:
			not_found.append(name)
			pkg_names.remove(name)
			continue

		pkg = cache[pkg_name]
		found = False
		for ver in pkg.versions:
			if ver.version == version:
				pkg.candidate = ver
				pkg_names.remove(name)
				pkg_names.append(pkg_name)
				found = True
				# Break because the same version could exist multiple times
				# Example, nala is in both Sid and Volian repository
				break

		if found:
			continue
		failed = True
		eprint(
			_("{error} Version {version} not found for package {package}").format(
				error=ERROR_PREFIX,
				version=color_version(version),
				package=color(pkg_name, "GREEN"),
			)
		)
	return not_found, failed


# pylint: disable=import-outside-toplevel, cyclic-import
def check_state(cache: Cache, nala_pkgs: PackageHandler) -> None:
	"""Check if pkg needs to be configured so we can show it."""
	from nala.nala import _fix_broken

	if cache.broken_count and arguments.fix_broken:
		_fix_broken(cache)
		sys.exit()
	for raw_pkg in cache._cache.packages:
		if raw_pkg.current_state in (
			CurrentState.HALF_CONFIGURED,
			CurrentState.UNPACKED,
		):
			pkg = cache[raw_pkg.name]
			if pkg.installed:
				nala_pkgs.configure_pkgs.append(
					NalaPackage(
						pkg.name, pkg.installed.version, pkg.installed.installed_size
					)
				)


def get_extra_pkgs(  # pylint: disable=too-many-branches
	extra_type: str,
	pkgs: Sequence[Package | NalaDebPackage],
	npkg_list: list[NalaPackage | list[NalaPackage]],
) -> None:
	"""Get Recommended or Suggested Packages."""
	dprint(f"Getting `{extra_type}` Packages")
	or_name = []
	for pkg in pkgs:
		if isinstance(pkg, Package):
			recommends: list[Dependency] | list[NalaDep]
			if not pkg.marked_install or not pkg.candidate:
				continue
			if not (recommends := pkg.candidate.get_dependencies(extra_type)):
				continue
		# Then The package must be a NalaDebPackage
		elif not (recommends := pkg.get_dependencies(extra_type)):
			continue

		for dep in recommends:
			# We don't need to show this if the extra is satisfied
			if dep.installed_target_versions:
				continue
			if len(dep) == 1:
				if not dep.target_versions:
					npkg_list.append(NalaPackage(dep[0].name, _("Virtual Package"), 0))
					continue
				ver = dep.target_versions[0]
				# We don't need to show if it's to be installed
				if ver.package.marked_install:
					continue
				npkg_list.append(NalaPackage(ver.package.name, ver.version, ver.size))
				continue
			or_deps = []
			for base_dep in dep:
				if not base_dep.target_versions:
					if base_dep.name in or_name:
						continue
					or_name.append(base_dep.name)
					or_deps.append(NalaPackage(base_dep.name, _("Virtual Package"), 0))
					continue
				ver = base_dep.target_versions[0]
				# We don't need to show if it's to be installed
				if ver.package.name in or_name or ver.package.marked_install:
					continue
				or_name.append(ver.package.name)
				or_deps.append(NalaPackage(ver.package.name, ver.version, ver.size))
			if len(or_deps) == 1:
				npkg_list.extend(or_deps)
				continue
			if or_deps:
				npkg_list.append(or_deps)


def check_broken(
	pkg_names: list[str], cache: Cache, remove: bool = False
) -> tuple[list[Package], list[str], bool]:
	"""Check if packages will be broken."""
	broken_count = 0
	broken: list[Package] = []
	depcache = cache._depcache

	not_found, failed = set_candidate_versions(pkg_names, cache)
	with cache.actiongroup():  # type: ignore[attr-defined]
		for pkg_name in pkg_names[:]:
			if pkg_name not in cache:
				not_found.append(pkg_name)
				continue

			pkg = cache[pkg_name]
			mark_pkg(pkg, depcache, remove=remove)
			if depcache.broken_count > broken_count and arguments.fix_broken:
				broken.append(pkg)
				broken_count += 1
	return broken, not_found, failed


def mark_pkg(pkg: Package, depcache: DepCache, remove: bool = False) -> bool:
	"""Mark Packages in depcache for broken checks."""
	if remove:
		if not pkg.installed and not (pkg.has_config_files and arguments.is_purge()):
			eprint(
				_("{notice} {package} is not installed").format(
					notice=NOTICE_PREFIX, package=color(pkg.name, "YELLOW")
				)
			)
			return False
		depcache.mark_delete(pkg._pkg, arguments.is_purge())
		return True

	# Check the installed version against the candidate version in case we're downgrading or upgrading.
	if (
		pkg.installed
		and pkg.candidate
		and pkg.installed.version == pkg.candidate.version
	):
		print(
			_("{package} is already at the latest version {version}").format(
				package=color(pkg.name, "GREEN"),
				version=color(pkg.installed.version, "BLUE"),
			)
		)
		return False
	depcache.mark_install(pkg._pkg, False, True)
	return True


def sort_pkg_changes(pkgs: list[Package], nala_pkgs: PackageHandler) -> None:
	"""Sort a list of packages and splits them based on the action to take."""
	dprint("Sorting Package Changes")
	for pkg in pkgs:
		installed = get_pkg_version(pkg, inst_first=True)
		if pkg.marked_delete:
			delete, autoremove = (
				(nala_pkgs.delete_pkgs, nala_pkgs.autoremove_pkgs)
				if pkg.installed
				else (nala_pkgs.delete_config, nala_pkgs.autoremove_config)
			)
			npkg = NalaPackage(pkg.name, installed.version, installed.installed_size)
			if pkg.name in nala_pkgs.autoremoved:
				autoremove.append(npkg)
				continue
			delete.append(npkg)

		candidate = get_pkg_version(pkg, cand_first=True)
		npkg = NalaPackage(pkg.name, candidate.version, candidate.size)
		if pkg.marked_install:
			nala_pkgs.install_pkgs.append(npkg)
			continue

		if pkg.marked_reinstall:
			nala_pkgs.reinstall_pkgs.append(npkg)
			continue

		npkg = NalaPackage(
			pkg.name, candidate.version, candidate.size, installed.version
		)
		if pkg.marked_upgrade:
			nala_pkgs.upgrade_pkgs.append(npkg)
			continue

		if pkg.marked_downgrade:
			nala_pkgs.downgrade_pkgs.append(npkg)
			continue


def need_reboot() -> bool:
	"""Check if the system needs a reboot and notify the user."""
	if REBOOT_REQUIRED.exists():
		if REBOOT_PKGS.exists():
			print(
				_("{notice} The following packages require a reboot.").format(
					notice=NOTICE_PREFIX
				)
			)

			for pkg in REBOOT_PKGS.read_text(encoding="utf-8").splitlines():
				print(f"  {color(pkg, 'GREEN')}")
			return False
		return True
	return False


def print_notices(notices: Iterable[str]) -> None:
	"""Print notices from dpkg."""
	if notices:
		print("\n" + color(_("Notices:"), "YELLOW"))
		for notice_msg in notices:
			if "NOTICE:" in notice_msg:
				notice_msg = notice_msg.replace("NOTICE:", NOTICE_PREFIX)
			if "Warning:" in notice_msg:
				notice_msg = notice_msg.replace("Warning:", WARNING_PREFIX)
			print(f"  {notice_msg}")


def setup_cache() -> Cache:
	"""Update the cache if necessary, and then return the Cache."""
	try:
		if arguments.update:
			with DelayedKeyboardInterrupt():
				with DpkgLive(install=False) as live:
					Cache().update(UpdateProgress(live))
	except (LockFailedException, FetchFailedException, apt_pkg.Error) as err:
		apt_error(err, arguments.command == "update")
	except KeyboardInterrupt:
		eprint(_("Exiting due to SIGINT"))
		sys.exit(ExitCode.SIGINT)
	except BrokenPipeError:
		sys.stderr.close()
	try:
		cache = Cache(OpProgress())
	except apt_pkg.Error as err:
		apt_error(err, True)
	return cache


def sort_pkg_name(pkg: Package) -> str:
	"""Sort by package name.

	This is to be used as sorted(key=sort_pkg_name)
	"""
	return f"{pkg.name}"


def check_term_ask() -> None:
	"""Check terminal and ask user if they want to continue."""
	# If we're piped or something the user should specify --assume-yes
	# As They are aware it can be dangerous to continue
	if (
		not term.console.is_terminal
		and not arguments.assume_yes
		and not arguments.assume_no
	):
		sys.exit(
			_(
				"{error} It can be dangerous to continue without a terminal. Use `--assume-yes`"
			).format(error=ERROR_PREFIX)
		)

	if not arguments.fix_broken:
		print(
			_("{warning} Using {switch} can be very dangerous!").format(
				warning=WARNING_PREFIX, switch=color("--no-fix-broken", "YELLOW")
			)
		)

	if not ask(_("Do you want to continue?")):
		eprint(_("Abort."))
		sys.exit(1)


def check_work(pkgs: list[Package], nala_pkgs: PackageHandler, operation: str) -> None:
	"""Check if there is any work for nala to do.

	Returns None if there is work, exit's successful if not.
	"""
	if nala_pkgs.configure_pkgs:
		return
	if operation == "upgrade" and not pkgs:
		print(color(_("All packages are up to date.")))
		sys.exit(0)
	elif operation == "install" and not pkgs and not nala_pkgs.local_debs:
		print(color(_("Nothing for Nala to do.")))
		sys.exit(0)
	elif operation == "remove" and not pkgs:
		print(color(_("Nothing for Nala to remove.")))
		sys.exit(0)
	elif operation == "fix-broken" and not pkgs:
		print(color(_("Nothing for Nala to fix.")))
		sys.exit(0)


def check_essential(pkgs: list[Package]) -> None:
	"""Check removal of essential packages."""
	dprint(f"Checking Essential: {not arguments.remove_essential}")
	if arguments.remove_essential:
		return
	essential: list[Text] = []
	for pkg in pkgs:
		if pkg.is_installed:
			# do not allow the removal of essential or required packages
			if pkg.essential and pkg.marked_delete:
				essential.append(from_ansi(color(pkg.name, "RED")))
			# do not allow the removal of nala
			elif pkg.shortname in "nala" and pkg.marked_delete:
				essential.append(from_ansi(color("nala", "RED")))

	if essential:
		essential_error(essential)



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Module for Nala debfile and dependency subclasses."""

from __future__ import annotations

import contextlib
from typing import List, cast

import apt_pkg
from apt.debfile import DebPackage
from apt.package import Version, VersionList

from nala.cache import Cache


class NalaDebPackage(DebPackage):
	"""A subclass for DebPackage to have attributes similar to Package."""

	def __init__(self, filename: str, cache: Cache) -> None:
		"""Subclass for DebPackage."""
		super().__init__(filename, cache)
		self.filename = filename
		self._cache = cache

	@property
	def name(self) -> str:
		"""Return the package name. This may have :arch appended."""
		return self.pkgname

	def _get_depends(self, _type: str) -> list[list[tuple[str, str, str]]]:
		"""List of packages on which this package depends on."""
		depends = []
		with contextlib.suppress(KeyError):
			depends.extend(apt_pkg.parse_depends(self._sections[_type], False))
		return depends

	def get_dependencies(self, *types: str) -> list[NalaDep]:
		"""Return a list of Dependency objects for the given types.

		Multiple types can be specified. Possible types are:
		'Breaks', 'Conflicts', 'Depends', 'Enhances', 'PreDepends',
		'Recommends', 'Replaces', 'Suggests'

		Additional types might be added in the future.
		"""
		depends_list = []
		for _type in types:
			for dep in self._get_depends(_type):
				base_deps = [
					NalaBaseDep(*dep_or, _type, cast(Cache, self._cache))
					for dep_or in dep
				]
				depends_list.append(NalaDep(base_deps, _type))
		return depends_list

	def installed_size(self) -> int:
		"""Get the installed size. Returns 0 if it cannot be found."""
		with contextlib.suppress(KeyError):
			return int(self._sections["Installed-Size"])
		return 0

	@property
	def dependencies(self) -> list[NalaDep]:
		"""Return the dependencies of the package version."""
		return self.get_dependencies("PreDepends", "Depends")


class NalaBaseDep:
	"""Base Dependency class to contain debfile deps."""

	def __init__(  # pylint: disable=too-many-arguments
		self,
		name: str,
		version: str,
		relation: str,
		rawtype: str,
		cache: Cache,
	) -> None:
		"""Initialize Base Dependency class to contain debfile deps."""
		self.name = name
		self.relation = relation
		self.relation_deb = relation
		self.version = version
		self.rawtype = rawtype
		self.cache = cache

	def __repr__(self) -> str:
		"""Return a string representation of the instance."""
		return (
			f"<NalaBaseDep: name:{self.name} relation:{self.relation}"
			f" version:{self.version} rawtype:{self.rawtype}>"
		)

	@property
	def rawstr(self) -> str:
		"""Return a string representation of the dependency.

		Returns the string representation of the dependency as it would be
		written in the debian/control file. The string representation does not
		include the type of the dependency.

		Example for an unversioned dependency:
		        python3

		Example for a versioned dependency:
		        python3 >= 3.2

		.. versionadded:: 1.0.0
		"""
		if self.version:
			return f"{self.name} {self.relation} {self.version}"
		return self.name

	@property
	def target_versions(self) -> VersionList:
		"""Return the target versions if they exist."""
		try:
			pkg = self.cache[self.name]
		except KeyError:
			return cast(VersionList, [])
		return pkg.versions

	@property
	def installed_target_versions(self) -> list[Version]:
		"""Return the installed target versions if they exist."""
		return [ver for ver in self.target_versions if ver.is_installed]


# Nuitka doesn't seem to like typing with the lower case list on this class.
class NalaDep(List[NalaBaseDep]):
	"""Dependency class to contain debfile deps."""

	def __init__(self, base_deps: list[NalaBaseDep], rawtype: str) -> None:
		"""Dependency class to contain debfile deps."""
		super().__init__(base_deps)
		self.rawtype = rawtype

	def __repr__(self) -> str:
		"""Return a string representation of the instance."""
		return f"<NalaDep: [{', '.join(repr(base_dep) for base_dep in self)}]>"

	@property
	def rawstr(self) -> str:
		"""Return a string representation of the Or-group of dependencies.

		Returns the string representation of the Or-group of dependencies as it
		would be written in the debian/control file.  The string representation
		does not include the type of the Or-group of dependencies.

		Example:
		-------
		        python2 >= 2.7 | python3

		"""
		return " | ".join(bd.rawstr for bd in self)

	@property
	def target_versions(self) -> list[Version]:
		"""List all Version objects which satisfy this Or-group of deps."""
		tvers: set[Version] = set()
		for base_dep in self:
			for tver in base_dep.target_versions:
				tvers.add(tver)
		return list(tvers)

	@property
	def installed_target_versions(self) -> list[Version]:
		"""List all Version installed Version objects which satisfy this dep."""
		return [ver for ver in self.target_versions if ver.is_installed]



#!/usr/bin/env python3
#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
# nala-cli.py doesn't conform to snake name
# This isn't really a module and is an entry script
# pylint: disable=invalid-name
"""Nala command-line entry point."""

from nala import __main__

__main__.main()



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Rich options for Nala output."""
from __future__ import annotations

import sys
from datetime import timedelta
from typing import Literal, cast

from rich import filesize
from rich.ansi import AnsiDecoder
from rich.box import Box
from rich.columns import Columns
from rich.console import Console, Group, RenderableType
from rich.live import Live, _RefreshThread
from rich.markup import escape
from rich.panel import Panel
from rich.pretty import Pretty
from rich.progress import (
	BarColumn,
	DownloadColumn,
	Progress,
	SpinnerColumn,
	Task,
	TaskID,
	TextColumn,
	TimeRemainingColumn,
	TransferSpeedColumn,
)
from rich.spinner import Spinner
from rich.style import Style
from rich.table import Column, Table
from rich.text import Text
from rich.tree import Tree

from nala import _, console
from nala.options import arguments

__all__ = (
	"Spinner",
	"Table",
	"Column",
	"Columns",
	"Console",
	"Tree",
	"Live",
	"Text",
	"escape",
	"Group",
	"TaskID",
	"Panel",
	"Pretty",
	"Progress",
	"RenderableType",
)


class Thread(_RefreshThread):
	"""A thread that calls refresh() at regular intervals.

	Subclass to change live.refresh with live.update.
	"""

	def run(self) -> None:
		while not self.done.wait(1 / self.refresh_per_second):
			with self.live._lock:
				if not self.done.is_set():
					self.live.scroll_bar(rerender=True)  # type: ignore[attr-defined]


def to_str(
	size: int,
	base: int,
) -> str:
	"""Format transfer speed to a string."""
	if arguments.config.get_bool("transfer_speed_bits", False):
		single = "bits"
		suffixes = ("Kbit", "Mbit", "Gbit", "Tbit", "Pbit", "Ebit", "Zbit", "Ybit")
		multiplier = 8
	else:
		single = "bytes"
		suffixes = ("KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
		multiplier = 1

	if size == 1:
		return f"1 {single}"
	if size < base:
		return f"{size:,} {single}"

	for i, suffix in enumerate(suffixes, 2):
		unit = base**i
		if size < unit:
			break

	display = (base * size / unit) * multiplier
	return f"{display:,.1f} {suffix}"


# pylint: disable=too-few-public-methods
class NalaTransferSpeed(TransferSpeedColumn):  # type: ignore[misc]
	"""Subclass of TransferSpeedColumn."""

	def render(self, task: Task) -> Text:
		"""Show data transfer speed."""
		if (speed := task.speed) is None:
			return Text("?", style="progress.data.speed")
		return Text(f"{to_str(int(speed), 1000)}/s", style="bold blue")


class NalaDownload(DownloadColumn):  # type: ignore[misc]
	"""Subclass of DownloadColumn."""

	def render(self, task: Task) -> Text:
		"""Calculate common unit for completed and total."""
		completed = int(task.completed)
		total = int(cast(float, task.total))  # type: ignore[redundant-cast]

		if arguments.config.get_bool("filesize_binary", False):
			unit, suffix = filesize.pick_unit_and_suffix(
				total,
				["bytes", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"],
				1024,
			)
		else:
			unit, suffix = filesize.pick_unit_and_suffix(
				total,
				["bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"],
				1000,
			)

		completed_ratio = completed / unit
		total_ratio = total / unit
		precision = 0 if unit == 1 else 1
		completed_str = f"{completed_ratio:,.{precision}f}"
		total_str = f"{total_ratio:,.{precision}f}"
		download_status = f"{completed_str}/{total_str} {suffix}"
		return Text(download_status, style="bold green")


class TimeRemaining(TimeRemainingColumn):  # type: ignore[misc]
	"""Renders estimated time remaining."""

	def render(self, task: Task) -> Text:
		"""Show time remaining."""
		remaining = task.time_remaining
		if remaining is None:
			return Text("-:--:--", style="bold default")
		remaining_delta = timedelta(seconds=int(remaining))
		return Text(f"{remaining_delta}", style="")


bar_back_style = Style(color="red")
bar_style = Style(color="cyan")
# Perform checks to see if we need to fall back to ascii.
is_utf8 = sys.stdout.encoding == "utf-8"
SEPARATOR = "[bold]•" if is_utf8 else "[bold]+"
SPIN_TYPE = "dots" if is_utf8 else "simpleDots"
FINISHED_TEXT = "[bold green]:heavy_check_mark:" if is_utf8 else " "
PROGRESS_PERCENT = "[bold blue]{task.percentage:>3.1f}%"
COMPLETED_TOTAL = "{task.completed}/{task.total}"
ELLIPSIS = "…" if is_utf8 else "..."
OVERFLOW = cast(Literal["crop"], "crop" if console.options.ascii_only else "ellipsis")

HORIZONTALS = Box(
	"\n".join(
		(
			"====",
			"    ",
			"====",
			"    ",
			"====",
			"====",
			"    ",
			"    ",
		)
	),
	ascii=True,
)

BAR_MAX = BarColumn(
	bar_width=None,
	# The background of our bar
	style=bar_back_style,
	# The color completed section
	complete_style=bar_style,
	# The color of completely finished bar
	finished_style=bar_style,
)
BAR_MIN = BarColumn(
	# The background of our bar
	style=bar_back_style,
	# The color completed section
	complete_style=bar_style,
	# The color of completely finished bar
	finished_style=bar_style,
)


def from_ansi(msg: str) -> Text:
	"""Convert ansi coded text into Rich Text."""
	return Text().join(AnsiDecoder().decode(msg))


def ascii_replace(string: str) -> str:
	"""If terminal is in ascii mode replace unicode characters."""
	return string if is_utf8 else string.encode("ascii", "replace").decode("ascii")


spinner = Spinner(SPIN_TYPE, style="bold blue")
time_remain = _("Time Remaining:")
pkg_download_progress = Progress(
	TextColumn(f"[bold green]{time_remain}"),
	TimeRemaining(),
	BAR_MAX,
	PROGRESS_PERCENT,
	SEPARATOR,
	NalaDownload(),
	SEPARATOR,
	NalaTransferSpeed(),
)
running_dpkg = _("Running dpkg")
dpkg_progress = Progress(
	SpinnerColumn(SPIN_TYPE, style="bold default", finished_text=FINISHED_TEXT),
	TextColumn(f"[bold blue]{running_dpkg} {ELLIPSIS}", justify="right"),
	BAR_MAX,
	PROGRESS_PERCENT,
	SEPARATOR,
	TimeRemaining(),
	SEPARATOR,
	COMPLETED_TOTAL,
)
testing = _("Testing Mirrors")
fetch_progress = Progress(
	# 	SpinnerColumn(SPIN_TYPE, style="bold blue"),
	TextColumn(f"[bold blue]{testing} {ELLIPSIS}", justify="right"),
	BAR_MIN,
	PROGRESS_PERCENT,
	SEPARATOR,
	COMPLETED_TOTAL,
	transient=True,
)



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2010 - 2021 Tatsuhiro Tsujikawa
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
# nala is based upon apt-metalink https://github.com/tatsuhiro-t/apt-metalink
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Module for package downloader."""
from __future__ import annotations

import asyncio
import contextlib
import hashlib
import re
import shutil
import sys
from asyncio import AbstractEventLoop, CancelledError, gather, run, sleep
from collections import Counter
from dataclasses import dataclass
from errno import ENOENT
from functools import partial
from pathlib import Path
from signal import Signals  # pylint: disable=no-name-in-module #Codacy
from signal import SIGINT, SIGTERM
from typing import Generator, Iterable, List, Sequence, Union, cast

from anyio import open_file
from apt.package import Package, Version
from apt_pkg import Configuration, config
from httpx import (
	AsyncClient,
	AsyncHTTPTransport,
	ConnectError,
	ConnectTimeout,
	HTTPError,
	HTTPStatusError,
	RemoteProtocolError,
	RequestError,
	get,
)

from nala import _, __version__, color
from nala.constants import (
	ARCHIVE_DIR,
	ERRNO_PATTERN,
	ERROR_PREFIX,
	NOTICE_PREFIX,
	PARTIAL_DIR,
	WARNING_PREFIX,
)
from nala.error import ExitCode, FileDownloadError
from nala.options import arguments
from nala.rich import (
	ELLIPSIS,
	Live,
	Panel,
	Pretty,
	Table,
	from_ansi,
	pkg_download_progress,
)
from nala.utils import dprint, eprint, get_pkg_name, term, unit_str, vprint

MIRROR_PATTERN = re.compile(r"mirror://(.*?/.*?)/")
MIRROR_FILE_PATTERN = re.compile(r"mirror\+file:(/.*?)/pool")
URL_PATTERN = re.compile(r"(https?://.*?/.*?)/")

STARTING_DOWNLOADS = color(_("Starting Downloads") + ELLIPSIS, "BLUE")

STARTING_DOWNLOAD = color(_("Starting Download:"), "BLUE")
# NOTE: "Starting Download: http://deb.debian.org/debian 3MB"
# NOTE: "Starting Download:" will be colored blue
STARTING_DOWNLOAD_STATUS = _("{starting_download} {url} {size}")

DOWNLOAD_COMPLETE = color(_("Download Complete:"), "GREEN")
# NOTE: "Download Complete: http://deb.debian.org/debian"
# NOTE: "Download Complete:" will be colored green
DOWNLOAD_COMPLETE_STATUS = _("{download_complete} {url}")

TOTAL_PACKAGES = color(_("Total Packages:"), "GREEN")
# NOTE: "Total Packages: 30/100"
# NOTE: "Starting Download:" will be colored green
TOTAL_PACKAGES_STATUS = _("{total_packages} {current}/{total}")

LAST_COMPLETED = color(_("Last Completed:"), "GREEN")
LAST_COMPLETED_STATUS = _("{last_completed} {package}")


REMOVING_FILE = _("{notice} Nala has removed {filename}")
FAILED_MOVE = _(
	"{error} Failed to move archive file, {str_err}: '{file1}' -> '{file2}'"
)
FILE_NO_EXIST = _("{error} {filename} does not exist!")
HASH_MISMATCH = _(
	"{error} Hash Sum does not match: {filename}\n"
	"  Expected Hash: {expected}\n"
	"  Received Hash: {received}"
)
SIZE_WRONG = _(
	"{error} File has unexpected size: {filename}\n"
	"  Expected Size: {expected}\n"
	"  Received Size: {received}"
)

## Debugging messages, No translation
HASH_STATUS = (
	"Hash Status = [\n    File: {filepath},\n"
	"    Candidate Hash: {hash_type} {expected},\n"
	"    Local Hash: {received},\n"
	"    Hash Success: {result},\n]"
)
HTTPX_STATUS_ERROR = (
	"Hash Status = [\n    File: {filepath},\n"
	"    Candidate Hash: {hash_type} {expected},\n"
	"    Local Hash: {received},\n"
	"    Hash Success: {result},\n]"
)
PEER_CLOSED = "peer closed connection without sending complete message body"
DownloadErrorTypes = Union[
	HTTPError, HTTPStatusError, RequestError, OSError, ConnectError, FileDownloadError
]


@dataclass
class URL:  # pylint: disable=too-many-instance-attributes
	"""Representation of a URL and File for download."""

	uri: str
	size: int
	path: Path
	proto: str = ""
	hash_type: str = "sha256"
	hash: str = ""
	failed: bool = False
	no_hash: bool = False

	def filename(self) -> str:
		"""Return the final portion of the filename."""
		return self.path.name

	def dprint(self, received: str) -> None:
		"""Debug print the URL's hash status."""
		dprint(
			HASH_STATUS.format(
				filepath=self.path,
				hash_type=self.hash_type.upper(),
				expected=self.hash or "Not Provided",
				received=received,
				result=received == self.hash,
			)
		)

	@staticmethod
	def from_version(version: Version) -> URL:
		"""Return a URL from an Apt Version."""
		return URL.new(version.uri or "", version)

	@staticmethod
	def new(uri: str, version: Version) -> URL:
		"""Return a Specific URL with an Apt Version."""
		return URL(
			uri,
			version.size,
			# Have to run the filename through a path to get the last section
			ARCHIVE_DIR / get_pkg_name(version),
			"",
			*get_hash(version),
		)


class URLSet(List[URL]):
	"""Set of urls that are all expected to provide the same file."""

	def size(self) -> int:
		"""Return the 'on disk size' for the completed download."""
		return self[0].size

	def filename(self) -> str:
		"""Return the final portion of the filename."""
		return self[0].filename()

	def path(self) -> Path:
		"""Return the destinateion Path object."""
		return self[0].path

	def any_available(self) -> bool:
		"""Return True if there are any urls that haven't failed."""
		return any(not url.failed for url in self)

	def next_available(self) -> URL | None:
		"""Return URLs that haven't failed."""
		return next((url for url in self if not url.failed), None)

	@staticmethod
	def from_version(version: Version) -> URLSet:
		"""Return a URLSet from an Apt Version."""
		url_set = URLSet()
		for uri in version.uris:
			url_set.append(
				URL(
					uri,
					version.size,
					ARCHIVE_DIR / get_pkg_name(version),
					*get_hash(version),
				)
			)
		return url_set

	@staticmethod
	def from_str(uri: str) -> URLSet:
		"""Return a URLSet from str uri."""
		url_set = URLSet()
		hash_type = hashsum = ""
		# This means a hash must have been specified
		if len(split_url := uri.split(":")) > 2:
			if len(split_url) != 4:
				sys.exit(
					_(
						"{error} Improper Hash Syntax\n"
						"  Try '{url}"
						":sha256:510b0c4f8fc3e7bd7182b53c0b81c1a113bea9fd3c18219eac0e18e601dc8d17'"
					).format(error=ERROR_PREFIX, url=f"{split_url[0]}:{split_url[1]}")
				)
			proto, url, hash_type, hashsum = split_url
			# Rebuild the uri for downloading
			uri = f"{proto}:{url}"

		# We must get the response so we know what the filesize is.
		response = get(uri, follow_redirects=True)
		response.raise_for_status()
		dprint(response.headers)
		try:
			size = int(response.headers["content-length"])
		except KeyError:
			sys.exit(
				_(
					"{error} No content length in response from {url}\n"
					"  Ensure the URL points to a Debian Package"
				).format(error=ERROR_PREFIX, url=uri)
			)
		path = ARCHIVE_DIR / Path(uri).name
		if hash_type and hashsum:
			url_set.append(URL(uri, size, path, hash_type, hashsum))
		else:
			url_set.append(URL(uri, size, path, no_hash=True))
		return url_set


class Downloader:  # pylint: disable=too-many-instance-attributes
	"""Manage Package Downloads."""

	def __init__(self, pkgs: Sequence[URLSet]) -> None:
		"""Manage Package Downloads."""
		dprint("Downloader Initializing")
		self.total_pkgs: int = len(pkgs)
		self.total_data: int = 0
		self.count: int = 0
		self.live: Live
		self.last_completed: str = ""
		self.proxy: dict[str, AsyncHTTPTransport] = {}
		self.failed: list[str] = []
		self.current: Counter[str] = Counter()
		self.fatal: bool = False
		self.exit: int | bool = False

		self.pkg_urls: Sequence[URLSet] = pkgs

		self.task = pkg_download_progress.add_task(
			"", total=sum(url.size() for url in pkgs)
		)

		self._set_proxy()
		dprint("Initialization Complete")

	def interrupt(self, signal_enum: Signals, loop: AbstractEventLoop) -> None:
		"""Shutdown the loop."""
		self.exit = 128 + signal_enum.real
		if self.exit == ExitCode.SIGINT:
			term.write(term.CURSER_UP + term.CLEAR_LINE)

		self.live.stop()
		for task in asyncio.all_tasks(loop):
			task.cancel()
		eprint(_("Exiting due to {signal}").format(signal=signal_enum.name))

	def _set_proxy(self) -> None:
		"""Set proxy configuration."""
		for proto in ("http", "https"):
			# This code is for the common cases of configuring proxies
			if common_proxy := config.find(f"Acquire::{proto}::Proxy"):
				# If the proxy is set to direct or false we disable it
				if common_proxy.lower() not in ("direct", "false"):
					self.proxy[f"{proto}://"] = AsyncHTTPTransport(proxy=common_proxy)  # type: ignore[arg-type]

			# The remainder of code is for proxying specific repos. Such a configuration may look like
			# Acquire::http::Proxy::deb.volian.org "xxx:8087"
			try:
				proxy_config = cast(
					Configuration, config.subtree(f"Acquire::{proto}::Proxy")  # type: ignore[attr-defined]
				)
			except KeyError:
				continue

			for key in proxy_config.keys():
				value: str | None = proxy_config.find(key)
				if not value:
					continue
				if "socks5h" in value or "socks5" in value:
					try:
						import socksio  # pylint: disable=import-outside-toplevel, unused-import

						value = value.replace("socks5h", "socks5")
					except ImportError:
						eprint(
							_(
								"{error} Using SOCKS proxy, but the 'socksio' package is not installed."
							).format(error=ERROR_PREFIX)
						)
						sys.exit(
							_(
								"{error} Install using 'nala install python3-socksio'"
							).format(error=ERROR_PREFIX)
						)
				# If direct or false, disable the proxy
				if value.lower() in ("direct", "false"):
					value = None
				self.proxy[f"{proto}://{key}"] = AsyncHTTPTransport(proxy=value)  # type: ignore[arg-type]

	async def _check_count(self, url: str) -> str:
		"""Check the url count and return if Nala should continue."""
		domain = ""
		if regex := URL_PATTERN.search(url):
			domain = regex.group(1)
			if self.current[domain] > 2:
				# Idk, but it doesn't work without the sleep.
				await sleep(0.01)
				return ""
			self.current[domain] += 1
		return domain

	def _gen_table(self) -> Panel:
		"""Generate Rich Table."""
		table = Table.grid()
		if arguments.debug:
			table.add_row(Pretty(self.current))
		table.add_row(
			from_ansi(
				TOTAL_PACKAGES_STATUS.format(
					total_packages=TOTAL_PACKAGES,
					current=self.count,
					total=self.total_pkgs,
				)
			)
		)

		if not self.last_completed:
			table.add_row(from_ansi(STARTING_DOWNLOADS))
		else:
			table.add_row(
				from_ansi(
					LAST_COMPLETED_STATUS.format(
						last_completed=LAST_COMPLETED, package=self.last_completed
					)
				)
			)

		table.add_row(pkg_download_progress.get_renderable())
		return Panel(
			table,
			title="[bold default]" + _("Downloading") + ELLIPSIS,
			title_align="left",
			border_style="bold green",
		)

	async def _update_progress(self, len_data: int, failed: bool = False) -> None:
		"""Update download progress."""
		if failed:
			self.total_data -= len_data
			pkg_download_progress.reset(self.task, completed=self.total_data)
			return
		self.total_data += len_data
		pkg_download_progress.advance(self.task, advance=len_data)
		self.live.update(self._gen_table())

	async def start_download(self) -> bool:
		"""Start async downloads."""
		if not self.pkg_urls:
			return True
		with Live(get_renderable=self._gen_table, refresh_per_second=10) as self.live:
			async with AsyncClient(
				timeout=20,
				mounts=self.proxy,
				follow_redirects=True,
				# Custom user agent fixes some downloading issues
				# Caused by httpx default agent sometimes being blocked.
				headers={"user-agent": f"nala/{__version__}"},
			) as client:
				loop = asyncio.get_running_loop()
				tasks = (
					loop.create_task(self._init_download(client, url))
					for url in self.pkg_urls
				)

				# Setup handlers for Interrupts
				for signal_enum in (SIGINT, SIGTERM):
					exit_func = partial(self.interrupt, signal_enum, loop)
					loop.add_signal_handler(signal_enum, exit_func)

				return all(await gather(*tasks))

	async def _init_download(self, client: AsyncClient, urls: URLSet) -> None:
		"""Download pkgs."""
		while url := urls.next_available():
			for url in urls:
				if not (domain := await self._check_count(url.uri)):
					continue
				try:
					await self._download(client, url)

					post_download_check(url)

					# Download completed. Raise count, lower domain counter
					self.count += 1
					self.current[domain] -= 1
					self.last_completed = url.filename()
					self.live.update(self._gen_table())
					return

				except (HTTPError, OSError, FileDownloadError) as error:
					url.failed = True
					self.current[domain] -= 1
					self.download_error(error, urls)
					continue

	async def _download(self, client: AsyncClient, url: URL) -> None:
		"""Download and write package."""
		vprint(
			STARTING_DOWNLOAD_STATUS.format(
				starting_download=STARTING_DOWNLOAD,
				url=url.uri,
				size=unit_str(url.size).strip(),
			)
		)

		dest = PARTIAL_DIR / url.filename()
		second_attempt = False
		while True:
			total_data = 0
			hash_fun = hashlib.new(url.hash_type)
			try:
				async with client.stream("GET", url.uri) as response:
					response.raise_for_status()
					async with await open_file(dest, mode="wb") as file:
						async for data in response.aiter_bytes():
							if data:
								await file.write(data)
								hash_fun.update(data)
								total_data += len(data)
								await self._update_progress(len(data))

			# Sometimes mirrors play a little dirty and close the connection
			except RemoteProtocolError as error:
				await self._update_progress(total_data, failed=True)
				if second_attempt:
					raise error from error
				second_attempt = True
				dprint(f"Mirror Failed: {url.uri} {error}, will try again.")
				continue

			url.dprint(received := hash_fun.hexdigest())
			# URL is no hash when local debs are downloaded without
			# Specifying a hash
			if url.no_hash:
				vprint(f"Skipping hashsum for {url.filename()} as one wasn't provided")
				break

			if url.hash != received:
				dest.unlink()
				self.fatal = True
				await self._update_progress(total_data, failed=True)
				raise FileDownloadError(
					errno=FileDownloadError.ERRHASH,
					filename=dest.name,
					expected=f"{url.hash_type.upper()}: {url.hash}",
					received=f"{url.hash_type.upper()}: {received}",
				)
			break

	def download_error(self, error: DownloadErrorTypes, urls: URLSet) -> None:
		"""Handle download errors."""
		print_error(error)

		if not (next_url := urls.next_available()):
			# No need to say there are no mirrors available in the event there is only one.
			if len(urls) > 1:
				eprint(
					_("{error} No more mirrors available for {filename}").format(
						error=ERROR_PREFIX, filename=color(urls.filename(), "YELLOW")
					)
				)

			# Status error are fatal as apt_pkg is likely to fail with these as well
			if isinstance(error, HTTPStatusError):
				# 401 Unauthenticated should be allowed to pass to apt for authentication
				if error.response.status_code == 401:
					return
				self.fatal = True

			self.failed.append(urls.filename())
			return

		eprint(
			_("{notice} Trying next url: {url}").format(
				notice=NOTICE_PREFIX, url=next_url.uri
			)
		)


def untrusted_error(untrusted: list[str]) -> None:
	"""Print the untrusted warnings and exit if we're not allowed."""
	eprint(
		_("{warning} The following packages cannot be authenticated!").format(
			warning=WARNING_PREFIX
		)
	)
	eprint(f"  {', '.join(untrusted)}")
	if not config.find_b("APT::Get::AllowUnauthenticated", False):
		sys.exit(
			_("{error} Some packages were unable to be authenticated").format(
				error=ERROR_PREFIX
			)
		)
	eprint(
		_(
			"{notice} Configuration is set to allow installation of unauthenticated packages."
		).format(notice=NOTICE_PREFIX)
	)


def print_error(error: DownloadErrorTypes) -> None:
	"""Print the download error to console."""
	if isinstance(error, ConnectTimeout):
		eprint(
			_("{error} {url} timed out").format(
				error=ERROR_PREFIX, url=error.request.url
			)
		)
		return
	if isinstance(error, ConnectError):
		# ConnectError: [Errno -2] Name or service not known
		errno_replace = re.sub(ERRNO_PATTERN, "", f"{error}").strip() + ":"
		eprint(f"{ERROR_PREFIX} {errno_replace} {error.request.url}")
		return
	if isinstance(error, FileDownloadError):
		file_error(error)
		return
	if isinstance(error, OSError) and error.errno != ENOENT:
		eprint(
			FAILED_MOVE.format(
				error=ERROR_PREFIX,
				str_err=error.strerror,
				file1=error.filename,
				file2=error.filename2,
			)
		)
		return

	# If 401 Unauthenticated we do not need to print anything
	# This is a secure repository like Ubuntu Pro. apt_pkg will pick it up
	if isinstance(error, HTTPStatusError) and error.response.status_code == 401:
		return

	msg = f"{error}" or type(error).__name__
	msg = msg.replace("\n", "\n  ")
	eprint(f"{ERROR_PREFIX} {msg}")


def file_error(error: FileDownloadError) -> None:
	"""Print the error from our FileDownloadError exception."""
	filename = color(error.filename, "YELLOW")
	if error.errno == FileDownloadError.ENOENT:
		eprint(FILE_NO_EXIST.format(error=ERROR_PREFIX, filename=filename))
		return
	if error.errno == FileDownloadError.ERRHASH:
		eprint(
			HASH_MISMATCH.format(
				error=ERROR_PREFIX,
				filename=filename,
				expected=error.expected,
				received=error.received,
			)
		)

	if error.errno == FileDownloadError.ERRSIZE:
		eprint(
			SIZE_WRONG.format(
				error=ERROR_PREFIX,
				filename=filename,
				expected=error.expected,
				received=error.received,
			)
		)

	eprint(REMOVING_FILE.format(notice=NOTICE_PREFIX, filename=filename))


def check_trusted(uri: str, candidate: Version) -> bool:
	"""Check if the candidate is trusted."""
	dprint(f"Checking trust of {candidate.package.name} ({candidate.version})")
	for (packagefile, _unused) in candidate._cand.file_list:
		if packagefile.site in uri and packagefile.archive != "now":
			indexfile = candidate.package._pcache._list.find_index(packagefile)
			dprint(f"{uri} = {indexfile and indexfile.is_trusted}")
			return bool(indexfile and indexfile.is_trusted)
	return False


def pre_download_check(url: URL) -> bool:
	"""Check if file exists, is correct, and run check hash."""
	dprint("Pre Download Package Check")

	if not url.path.exists():
		dprint(f"File Doesn't exist: {url.filename()}")
		return False

	if (size := url.path.stat().st_size) != url.size:
		dprint(f"File {url.filename()} has an unexpected size {size} != {url.size}")
		url.path.unlink()
		return False

	try:
		if not check_hash(url):
			dprint(f"Hash Checking has failed. Removing: {url.filename()}")
			url.path.unlink()
			return False
		dprint(f"Package doesn't require download: {url.filename()}")
		return True
	except OSError as err:
		eprint(_("Failed to check hash"), err)
		return False


def post_download_check(url: URL) -> bool:
	"""Check if file exists, is correct, and run check hash."""
	dprint("Post Download Package Check")

	source = PARTIAL_DIR / url.filename()
	try:
		dprint(f"Moving {source} -> {url.path}")
		source.rename(url.path)
	except OSError as error:
		if error.errno != ENOENT:
			eprint(
				FAILED_MOVE.format(
					error=ERRNO_PATTERN,
					str_err=error.strerror,
					file1=error.filename,
					file2=error.filename2,
				)
			)

	if not url.path.exists():
		dprint(f"File Doesn't exist: {url.filename()}")
		raise FileDownloadError(
			errno=FileDownloadError.ENOENT,
			filename=url.filename(),
		)

	if (size := url.path.stat().st_size) != url.size:
		dprint(f"File {url.filename()} has an unexpected size {size} != {url.size}")
		url.path.unlink()
		raise FileDownloadError(
			errno=FileDownloadError.ERRSIZE,
			filename=url.filename(),
			expected=f"{url.size}",
			received=f"{size}",
		)

	vprint(
		DOWNLOAD_COMPLETE_STATUS.format(
			download_complete=DOWNLOAD_COMPLETE, url=url.uri
		)
	)

	# Hash was checked while downloading. We can just call it good.
	return True


def check_hash(url: URL) -> bool:
	"""Check hash value."""
	hash_fun = hashlib.new(url.hash_type)
	with url.path.open("rb") as file:
		while data := file.read(4096):
			hash_fun.update(data)

	received = hash_fun.hexdigest()
	url.dprint(received)
	return received == url.hash


def get_hash(version: Version) -> tuple[str, str]:
	"""Get the correct hash value."""
	hash_list = version._records.hashes
	hashes = ("SHA512", "SHA256")

	# From Debian's requirements we are not to use these for security checking.
	# https://wiki.debian.org/DebianRepository/Format#MD5Sum.2C_SHA1.2C_SHA256
	# Clients may not use the MD5Sum and SHA1 fields for security purposes,
	# and must require a SHA256 or a SHA512 field.
	# hashes = ('SHA512', 'SHA256', 'SHA1', 'MD5')

	for _type in hashes:
		with contextlib.suppress(KeyError):
			return _type.lower(), hash_list.find(_type).hashvalue

	filename = Path(version.filename).name or version.package.name
	eprint(
		_("{error} {filename} can't be checked for integrity.").format(
			error=ERROR_PREFIX, filename=color(filename, "YELLOW")
		)
	)
	sys.exit(_("There are no hashes available for this package."))


def file_exists(cand: Version, uris: Iterable[str]) -> bool:
	"""Check if the file exists and move it if so, and hash check."""
	for uri in uris:
		if uri.startswith("file"):
			src = Path(uri.lstrip("file:"))
			dest = ARCHIVE_DIR / src.name
			# Make sure that the source exists and the destination does not
			if src.is_file() and not dest.is_file():
				# Move the file to the archive directory.
				# We're allowed to do this silently because hashsum comes later
				dprint(f"{src} => {shutil.copy2(src, dest)}")

				# pre_download_check needs to happen for hash checking
				if pre_download_check(URL.new(uri, cand)):
					return True
	return False


def filter_local_repo(pkgs: Iterable[Package]) -> list[URLSet]:
	"""Filter any local repository packages.

	This will check if the packages are coming from a local repository.

	If a package is from a local repo it will move it into the archive directory.

	Lastly it will checksum all packages that exist on the system
	and return a list of packages that need to be downloaded.
	"""
	untrusted: list[str] = []
	mirrors: dict[str, list[str]] = {}
	urls: list[URLSet] = []
	for pkg in pkgs:
		# At this point anything that makes it will have a candidate
		if not (cand := pkg.candidate) or pkg.marked_delete:
			continue

		uris = list(filter_uris(cand, mirrors))
		dprint(f"Filtered URIs: {uris}")
		# Check trust of all potentially used sources
		for uri in uris:
			if not check_trusted(uri, cand):
				untrusted.append(color(cand.package.name, "RED"))

		# Exit with an error if there are unauthenticated packages
		# This can proceed if overridden by configuration
		if untrusted:
			untrusted_error(untrusted)

		if file_exists(cand, uris):
			continue

		# Check for http uris to download
		url_set = URLSet()
		for uri in uris:
			if not uri.startswith("http"):
				continue

			url = URL.new(uri, cand)
			if not pre_download_check(url):
				url_set.append(url)

		if len(url_set) > 0:
			urls.append(url_set)

	# Return the list of packages that should be downloaded
	dprint(urls)
	return urls


def filter_uris(
	candidate: Version, mirrors: dict[str, list[str]]
) -> Generator[str, None, None]:
	"""Filter uris into usable urls."""
	for uri in candidate.uris:
		# Regex to check if we're using mirror://
		if regex := MIRROR_PATTERN.search(uri):
			set_mirrors_txt(domain := regex.group(1), mirrors)
			yield from (
				link + candidate.filename
				for link in mirrors[domain]
				if not link.startswith("#")
			)
			continue
		# Regex to check if we're using mirror+file:/
		if regex := MIRROR_FILE_PATTERN.search(uri):
			if regex.group(1) not in mirrors:
				# There are some other options in the mirror file
				# I don't believe it's necessary to implement them
				# So I am just splitting on whitespace.
				# See https://gitlab.com/volian/nala/-/issues/323
				mirrors[regex.group(1)] = discard_after_whitespace(
					Path(regex.group(1)).read_text(encoding="utf-8").splitlines()
				)

			yield from (
				f"{link}/{candidate.filename}"
				for link in mirrors[regex.group(1)]
				if not link.startswith("#")
			)
			continue
		yield uri


def discard_after_whitespace(lines: list[str]) -> list[str]:
	"""Split on whitespace and return the very first item."""
	return [line.split(maxsplit=1)[0] for line in lines]


def set_mirrors_txt(domain: str, mirrors: dict[str, list[str]]) -> None:
	"""Check if user has mirrors:// and handle accordingly."""
	if domain not in mirrors:
		url = f"http://{domain}"
		try:
			mirrors[domain] = discard_after_whitespace(
				get(url, follow_redirects=True).text.splitlines()
			)
		except HTTPError:
			sys.exit(
				_("{error} unable to connect to {url}").format(
					error=ERROR_PREFIX, url=url
				)
			)


def download(downloader: Downloader) -> None:
	"""Run downloads and check for failures.

	Does not return if in Download Only mode.
	"""
	try:
		run(downloader.start_download())
	except (CancelledError, RuntimeError) as error:
		if downloader.exit:
			sys.exit(downloader.exit)
		raise error from error
	except ValueError as error:
		sys.exit(f"{ERROR_PREFIX} {error}")

	if arguments.download_only and not downloader.failed:
		print(_("Download complete and in download only mode."))
		sys.exit(0)

	if downloader.failed:
		eprint(
			_("{error} Download failure. The following downloads failed:").format(
				error=ERROR_PREFIX
			)
		)
		eprint(f"  {', '.join(color(pkg, 'YELLOW') for pkg in downloader.failed)}")

		if downloader.fatal:
			sys.exit(1)
		if arguments.download_only:
			sys.exit(_("In download only mode. Not falling back to apt_pkg."))
		eprint(
			_(
				"{warning} Falling back to apt_pkg. The following downloads failed:"
			).format(warning=WARNING_PREFIX)
		)


def download_pkgs(pkgs: Iterable[Package]) -> None:
	"""Download package from a list of pkgs."""
	download(
		Downloader(
			# Start the larger files first, as they take the longest
			sorted(
				filter_local_repo(pkgs),
				key=lambda url: url.size(),
				reverse=True,
			)
		)
	)


def download_strings(urls: Iterable[str]) -> None:
	"""Download packages from a list of urls.

	This function assumes that each URL is its own package
	and will create a URLSet for each str url in the iterable
	"""
	download(Downloader([URLSet.from_str(url) for url in urls]))



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Nala dpkg module."""
from __future__ import annotations

import contextlib
import errno
import fcntl
import os
import pty
import re
import signal
import struct
import sys
import termios
from time import sleep, time
from traceback import format_exception
from types import FrameType
from typing import Match, TextIO

import apt_pkg
from apt.progress import base, text
from pexpect.fdpexpect import fdspawn
from pexpect.utils import poll_ignore_interrupts
from ptyprocess.ptyprocess import _setwinsize

from nala import _, color
from nala.constants import (
	DPKG_ERRORS,
	DPKG_STATUS,
	ERROR_PREFIX,
	HANDLER,
	NOTICES,
	SPAM,
	WARNING_PREFIX,
)
from nala.options import arguments
from nala.rich import (
	ELLIPSIS,
	OVERFLOW,
	Group,
	Live,
	Panel,
	RenderableType,
	Table,
	TaskID,
	Thread,
	ascii_replace,
	dpkg_progress,
	from_ansi,
	spinner,
	to_str,
)
from nala.utils import dprint, eprint, term, unit_str

VERSION_PATTERN = re.compile(r"\(.*?\)")
PARENTHESIS_PATTERN = re.compile(r"[()]")

notice: list[str] = []
pkgnames: set[str] = set()
unpacked: set[str] = set()
dpkg_error: list[str] = []
update_error: list[str] = []

REMOVING = "Removing"
UNPACKING = "Unpacking"
SETTING_UP = "Setting up"
PROCESSING = "Processing"
GET = "GET"

UPDATED = _("Updated:")
DOWNLOADED = _("Downloaded:")
IGNORED = _("Ignored:")
NO_CHANGE = _("No Change:")

# NOTE: Spacing of following status messages
# NOTE: is to allow the urls to be properly aligned
# NOTE: Especially if your status would come after the package
# NOTE: You do not have to follow this scheme
# NOTE: but do note that the headers will be colored regardless
# NOTE: No Change: http://deb.volian.org/volian scar InRelease
# NOTE: Ignored:   http://deb.volian.org/volian scar InRelease
# NOTE: Updated:   http://deb.volian.org/volian scar InRelease
NO_CHANGE_MSG = _("{no_change} {info}")
NO_CHANGE_SIZE_MSG = _("{no_change} {info} [{size}]")
IGNORED_MSG = _("{ignored}   {info}")
UPDATE_MSG = _("{updated}   {info}")
UPDATE_SIZE_MSG = _("{updated}   {info} [{size}]")

REMOVING_HEAD = color(_("Removing:"), "RED")
UNPACKING_HEAD = color(_("Unpacking:"), "GREEN")
SETTING_UP_HEAD = color(_("Setting up:"), "GREEN")
PROCESSING_HEAD = color(_("Processing:"), "GREEN")

# NOTE: Spacing of following status messages
# NOTE: is to allow dpkg messages to be properly aligned
# NOTE: Especially if your status would come after the package
# NOTE: You do not have to follow this scheme
# NOTE: but do note that the headers will be colored regardless
# NOTE: Unpacking:   neofetch (7.1.0-3)
# NOTE: Setting up:  neofetch (7.1.0-3)
# NOTE: Removing:    neofetch (7.1.0-3)
# NOTE: Processing:  triggers for man-db (2.10.2-1)
# NOTE: You can change the headers and positions as you would like,
# NOTE: but do note that the headers will be colored regardless
SETTING_UP_MSG = _("{setting_up} {dpkg_msg}")
PROCESSING_MSG = _("{processing} {dpkg_msg}")
UNPACKING_MSG = _("{unpacking}  {dpkg_msg}")
# NOTE: That's the end of alignment spacing
REMOVING_MSG = _("{removing}   {dpkg_msg}")

# NOTE: This translation is separate from the one below
# NOTE: Because we do a check specifically on this string
FETCHED = _("Fetched")
# NOTE: Fetched 81.0 MB in 6s (1448 kB/s)
FETCHED_MSG = _("{fetched} {size} in {elapsed} ({speed}/s)")


class OpProgress(text.OpProgress):
	"""Operation progress reporting.

	This closely resembles OpTextProgress in libapt-pkg.
	"""

	# we have to use this string format or else things get buggy
	# pylint: disable=consider-using-f-string
	def update(self, percent: float | None = None) -> None:
		"""Call periodically to update the user interface."""
		base.OpProgress.update(self, percent)
		if arguments.verbose:
			if self.major_change and self.old_op:
				self._write(self.old_op)

			self._write("%s... %i%%\r" % (self.op, self.percent), False, True)
			self.old_op = self.op

	def done(self) -> None:
		"""Call once an operation has been completed."""
		base.OpProgress.done(self)
		if arguments.verbose:
			if self.old_op:
				self._write(_("%c%s... Done") % ("\r", self.old_op), True, True)
			self.old_op = ""


class UpdateProgress(text.AcquireProgress):
	"""Class for getting cache update status and printing to terminal."""

	def __init__(self, live: DpkgLive) -> None:
		"""Class for getting cache update status and printing to terminal."""
		dprint("Init UpdateProgress")
		text.AcquireProgress.__init__(self)
		self._file = sys.__stdout__
		self._signal: HANDLER = None
		self._id = 1
		self._width = 80
		self.live = live
		self.elapsed = 0.0

		self.live.scroll_list.clear()

	def apt_write(self, msg: str, newline: bool = True, maximize: bool = False) -> None:
		"""Write original apt update message."""
		self._file.write("\r")
		self._file.write(msg)

		# Fill remaining stuff with whitespace
		if self._width > len(msg):
			self._file.write((self._width - len(msg)) * " ")
		elif maximize:  # Needed for OpProgress.
			self._width = max(self._width, len(msg))
		if newline:
			self._file.write("\n")
		else:
			self._file.flush()

	def _write(self, msg: str, newline: bool = True, maximize: bool = False) -> None:
		"""Write the message on the terminal, fill remaining space."""
		if arguments.raw_dpkg or not term.can_format():
			self.apt_write(msg, newline, maximize)
			return

		for item in (UPDATED, DOWNLOADED, IGNORED, NO_CHANGE):
			if item in msg:
				self.table_print(msg, update_spinner=True)
				break
		else:
			if FETCHED in msg:
				self.table_print(msg, fetched=True)
				return

			if ERROR_PREFIX in msg:
				for line in msg.splitlines():
					update_error.append(line)
					eprint(line)
				return

			spinner.text = from_ansi(msg)
			self.table_print(update_spinner=True)

	def table_print(
		self, msg: str = "", fetched: bool = False, update_spinner: bool = False
	) -> None:
		"""Update wrapper for the scroll bar."""
		if not arguments.scroll and not fetched and msg:
			print(msg)
			return

		self.live.scroll_bar(
			msg,
			update_spinner=update_spinner,
			apt_fetch=self.live.install,
			use_bar=False,
		)

	def ims_hit(self, item: apt_pkg.AcquireItemDesc) -> None:
		"""Call when an item is update (e.g. not modified on the server)."""
		base.AcquireProgress.ims_hit(self, item)
		line = NO_CHANGE_MSG.format(
			no_change=color(NO_CHANGE, "GREEN"), info=item.description
		)
		if item.owner.filesize:
			line = NO_CHANGE_SIZE_MSG.format(
				no_change=color(NO_CHANGE, "GREEN"),
				info=item.description,
				size=unit_str(item.owner.filesize).strip(),
			)
		self._write(line)

	def fail(self, item: apt_pkg.AcquireItemDesc) -> None:
		"""Call when an item is failed."""
		base.AcquireProgress.fail(self, item)
		if item.owner.status == item.owner.STAT_DONE:
			self._write(
				IGNORED_MSG.format(
					ignored=color(IGNORED, "YELLOW"), info=item.description
				)
			)
		else:
			# This doesn't need to be translated. Just an error dump
			self._write(f"{ERROR_PREFIX} {item.description}\n  {item.owner.error_text}")

	def fetch(self, item: apt_pkg.AcquireItemDesc) -> None:
		"""Call when some of the item's data is fetched."""
		base.AcquireProgress.fetch(self, item)
		# It's complete already (e.g. Hit)
		if item.owner.complete:
			return
		line = UPDATE_MSG.format(
			updated=color(DOWNLOADED if self.live.install else UPDATED, "BLUE"),
			info=item.description,
		)
		if item.owner.filesize:
			line = UPDATE_SIZE_MSG.format(
				updated=color(DOWNLOADED if self.live.install else UPDATED, "BLUE"),
				info=item.description,
				size=unit_str(item.owner.filesize).strip(),
			)
		self._write(line)

	def _winch(self, *_args: object) -> None:
		"""Signal handler for window resize signals."""
		if hasattr(self._file, "fileno") and os.isatty(self._file.fileno()):
			buf = fcntl.ioctl(self._file, termios.TIOCGWINSZ, 8 * b" ")
			dummy, columns, dummy, dummy = struct.unpack("hhhh", buf)
			self._width = columns - 1  # 1 for the cursor

	def start(self) -> None:
		"""Start an Acquire progress.

		In this case, the function sets up a signal handler for SIGWINCH, i.e.
		window resize signals. And it also sets id to 1.
		"""
		base.AcquireProgress.start(self)
		self.elapsed = time()
		self._signal = signal.signal(signal.SIGWINCH, self._winch)
		# Get the window size.
		self._winch()
		self._id = 1

	def final_msg(self) -> str:
		"""Print closing fetched message."""
		elapsed_time = int(time() - self.elapsed)
		if elapsed_time == 0:
			fetched_speed = self.current_cps
		else:
			fetched_speed = int(self.fetched_bytes) / elapsed_time
		return color(
			FETCHED_MSG.format(
				fetched=FETCHED,
				size=unit_str(int(self.fetched_bytes)).strip(),
				elapsed=apt_pkg.time_to_str(elapsed_time),
				speed=to_str(int(fetched_speed), 1000).strip(),
			)
		)

	def stop(self) -> None:
		"""Invoke when the Acquire process stops running."""
		base.AcquireProgress.stop(self)
		# We don't want to display fetched Zero if we're in apt fetch.
		if self.fetched_bytes != 0 or not self.live.install:
			self._write(self.final_msg())
		# Delete the signal again.
		signal.signal(signal.SIGWINCH, self._signal)

	def pulse(  # pylint: disable=too-many-branches
		self, owner: apt_pkg.Acquire
	) -> bool:
		"""Periodically invoked while the Acquire process is underway."""
		base.AcquireProgress.pulse(self, owner)
		# only show progress on a tty to not clutter log files etc
		if hasattr(self._file, "fileno") and not os.isatty(self._file.fileno()):
			return True

		# calculate progress
		percent = ((self.current_bytes + self.current_items) * 100.0) / (
			self.total_bytes + self.total_items
		)

		shown = False
		tval = f"{percent:.0f}%"
		end = ""
		if self.current_cps:
			eta = int((self.total_bytes - self.current_bytes) / self.current_cps)
			end = f" {unit_str(int(self.current_cps)).strip()}/s {apt_pkg.time_to_str(eta)}"

		for worker in owner.workers:
			val = ""
			if not worker.current_item:
				if worker.status:
					val = f" [{worker.status}]"
					if len(tval) + len(val) + len(end) >= self._width - 5:
						break
					tval += val
					shown = True
				continue
			shown = True

			if worker.current_item.owner.id:
				val += (
					f" [{worker.current_item.owner.id} {worker.current_item.shortdesc}"
				)
			else:
				val += f" [{worker.current_item.description}"
			if worker.current_item.owner.active_subprocess:
				val += f" {worker.current_item.owner.active_subprocess}"

			val += f" {unit_str(worker.current_size).strip()}"

			# Add the total size and percent
			if worker.total_size and not worker.current_item.owner.complete:
				val += (
					f"/{unit_str(worker.total_size).strip()}"
					f" {(worker.current_size * 100.0) / worker.total_size:.0f}%"
				)

			val += "]"

			if len(tval) + len(val) + len(end) >= self._width - 5:
				# Display as many items as screen width
				break
			tval += val

		if not shown:
			tval += _(" [Working]")

		if self.current_cps:
			tval += (self._width - len(end) - len(tval) - 5) * " " + end

		self._write(tval, False)
		return True


# pylint: disable=too-many-instance-attributes, too-many-public-methods, too-many-arguments, too-many-lines
class InstallProgress(base.InstallProgress):
	"""Class for getting dpkg status and printing to terminal."""

	def __init__(
		self,
		dpkg_log: TextIO,
		term_log: TextIO,
		live: DpkgLive,
		task: TaskID,
		config_purge: tuple[str, ...],
	) -> None:
		"""Class for getting dpkg status and printing to terminal."""
		dprint("Init InstallProgress")
		base.InstallProgress.__init__(self)
		self.task = task
		self._dpkg_log = dpkg_log
		self._term_log = term_log
		self.live = live
		self.config_purge = config_purge
		self.raw = False
		self.last_line = b""
		self.child: AptExpect
		self.child_fd: int
		self.child_pid: int
		# Setting environment to xterm seems to work fine for linux terminal
		# I don't think we will be supporting much more this this, at least for now
		if not term.is_xterm() and not arguments.raw_dpkg:
			os.environ["TERM"] = "xterm"

	def finish_update(self) -> None:
		"""Call when update has finished."""
		if not arguments.raw_dpkg:
			dpkg_progress.advance(self.task)
			self.live.scroll_bar()

	def run_install(self, apt: apt_pkg.PackageManager | list[str]) -> int:
		"""Install using the `PackageManager` object `obj`.

		returns the result of calling `obj.do_install()`
		"""
		dprint("Forking")
		pid, self.child_fd = fork()
		if pid == 0:
			try:
				# PEP-446 implemented in Python 3.4 made all descriptors
				# CLOEXEC, but we need to be able to pass writefd to dpkg
				# when we spawn it
				os.set_inheritable(self.writefd, True)
				if not isinstance(apt, apt_pkg.PackageManager):
					# pylint: disable=subprocess-run-check
					self.dpkg_log("Command Execution:\n")
					self.dpkg_log(f"Command = {apt}\n\n")
					os._exit(
						os.spawnlp(  # nosec
							os.P_WAIT,
							"dpkg",
							"dpkg",
							"--status-fd",
							f"{self.write_stream.fileno()}",
							"-i",
							*apt,
						)
					)
				# We ignore this with mypy because the attr is there
				self.dpkg_log("Apt Do Install\n\n")
				os._exit(apt.do_install(self.write_stream.fileno()))  # type: ignore[attr-defined]
			# We need to catch every exception here.
			# If we don't the code continues in the child,
			# And bugs will be very confusing
			except Exception:  # pylint: disable=broad-except
				exception = format_exception(*sys.exc_info())
				self.dpkg_log(f"{exception}\n")
				os._exit(1)

		dprint("Dpkg Forked")
		self.child_pid = pid
		if arguments.raw_dpkg:
			return os.WEXITSTATUS(self.wait_child())
		# We use fdspawn from pexpect to interact with our dpkg pty
		# But we also subclass it to give it the interact method and setwindow
		self.child = AptExpect(self.child_fd, timeout=None)

		signal.signal(signal.SIGWINCH, self.sigwinch_passthrough)
		self.child.interact(self)
		return os.WEXITSTATUS(self.wait_child())

	def sigwinch_passthrough(
		self, _sig_dummy: int, _data_dummy: FrameType | None
	) -> None:
		"""Pass through sigwinch signals to dpkg."""
		buffer = struct.pack("HHHH", 0, 0, 0, 0)
		term_size = struct.unpack(
			"hhhh", fcntl.ioctl(term.STDIN, termios.TIOCGWINSZ, buffer)
		)
		if self.child.isalive():
			with contextlib.suppress(ValueError):
				_setwinsize(self.child_fd, term_size[0], term_size[1])

	def dpkg_log(self, msg: str) -> None:
		"""Write to dpkg-debug.log and flush."""
		self._dpkg_log.write(msg)
		self._dpkg_log.flush()

	def term_log(self, msg: bytes) -> None:
		"""Write to term.log and flush."""
		self._term_log.write(f"{msg.decode('utf-8').strip()}\n")
		self._term_log.flush()

	def dpkg_status(self, data: bytes) -> bool:
		"""Handle any status messages."""
		for status in DPKG_STATUS:
			if status in data:
				statuses = data.split(b"\r")
				if len(statuses) > 2:
					self.dpkg_log(f"Status_Split = {repr(statuses)}\n")
				for msg in statuses:
					if msg != b"":
						spinner.text = from_ansi(color(msg.decode().strip()))
						self.live.scroll_bar(update_spinner=True)
				self.dpkg_log(term.LF.decode())
				return True
		return False

	def read_status(self) -> None:
		"""Read the status fd and send it to update progress bar."""
		try:
			status = self.status_stream.read(1024)
		except OSError as err:
			# Resource temporarily unavailable is ignored
			if err.errno not in (errno.EAGAIN, errno.EWOULDBLOCK):
				print(err.strerror)
			return
		for line in status.splitlines():
			self.update_progress_bar(line)

	def update_progress_bar(self, line: str) -> None:
		"""Update the interface."""
		pkgname = status = status_str = _percent = base_status = ""

		if line.startswith("pm"):
			try:
				(status, pkgname, _percent, status_str) = line.split(":", 3)
			except ValueError:
				# Silently ignore lines that can't be parsed
				return
		elif line.startswith("status"):
			try:
				(base_status, pkgname, status, status_str) = line.split(":", 3)
			except ValueError:
				(base_status, pkgname, status) = line.split(":", 2)

		# Always strip the status message
		pkgname = pkgname.strip()
		status_str = status_str.strip()
		status = status.strip()

		# This is the main branch for apt installs
		if status == "pmstatus":
			dprint(f"apt: {pkgname} {status_str}")
			if status_str.startswith(("Unpacking", "Removing")):
				unpacked.add(pkgname)
				self.advance_progress()
			# Either condition can satisfy this mark provided the package hasn't been advanced
			elif (
				status_str.startswith(("Installed", "Configuring"))
				and pkgname not in pkgnames
				and pkgname in unpacked
			):
				pkgnames.add(pkgname)
				self.advance_progress()
		# This branch only happens for local .deb installs.
		elif base_status == "status":
			# Sometimes unpacked is notified twice for one package
			# We check against out set to make sure not to over shoot progress
			if status == "unpacked" and pkgname not in pkgnames:
				self.advance_progress()
				pkgnames.add(pkgname)
			# Sometimes packages are notified as installed
			# But we only care for ones that have been unpacked
			if status == "installed" and pkgname in pkgnames:
				self.advance_progress()
			dprint(f"dpkg: {pkgname} {status}")

	def pre_filter(self, data: bytes) -> None:
		"""Filter data from interact."""
		self.read_status()

		# This is a work around for a hang in non-interactive mode
		# https://github.com/liske/needrestart/issues/129
		if (
			os.environ.get("DEBIAN_FRONTEND") == "noninteractive"
			and b"[Return]" in data
		):
			os.write(self.child_fd, term.CRLF)

		# Save Term and Alt Screen for debconf and Bracked Paste for the start of the shell
		if (
			term.SAVE_TERM in data
			or term.ENABLE_BRACKETED_PASTE in data
			or term.ENABLE_ALT_SCREEN in data
		):
			self.raw_init()

		self.dpkg_log(f"Raw = {self.raw}: [{repr(data)}]\n")

		if not self.raw and self.dpkg_status(data):
			return
		self.format_dpkg_output(data)

	def format_dpkg_output(self, rawline: bytes) -> None:
		"""Facilitate what needs to happen to dpkg output."""
		# If we made it here that means we're okay to start a new line in the log
		self.dpkg_log(term.LF.decode())

		if self.raw:
			self.rawline_handler(rawline)
			return

		line = ascii_replace(rawline.decode().strip())

		if check_line_spam(line, rawline, self.last_line):
			return

		if (
			self.config_purge
			and "Purging configuration files" in line
			and any(pkg in line for pkg in self.config_purge)
		):
			self.advance_progress()

		self.term_log(rawline)

		# We need to split up large lines and handle them individually.
		if rawline.count(b"\r\n") > 1:
			data_split = rawline.strip().split(b"\r\n")
			error = data_split[0].decode() in dpkg_error
			self.dpkg_log(f"Data_Split = {repr(data_split)}\n")
			for new_line in data_split:
				if not new_line:
					continue
				check_error(rawline, new_line.decode(), error)
				self.format_write(new_line.decode(), rawline)
			return
		self.format_write(line, rawline)

	def format_write(self, line: str, rawline: bytes) -> None:
		"""Format the line if and handle writing it to the terminal."""
		# Main format section for making things pretty
		msg = msg_formatter(line)
		# If verbose we just send it. No bars
		if not arguments.scroll:
			print(msg)
			self.live.scroll_bar()
		elif "Fetched:" in msg:
			# This is some magic for apt-listdifferences to put
			# the fetched message in the spinner since it gets spammy
			spinner.text = from_ansi(color(" ".join(line.split()[1:])))
			self.live.scroll_bar(msg, update_spinner=True)
		else:
			self.live.scroll_bar(msg)
		if (stdout := sys.stdout) is not None:
			stdout.flush()
		self.set_last_line(rawline)

	def rawline_handler(self, rawline: bytes) -> None:
		"""Handle text operations for rawline."""
		term.write(rawline)
		# Once we write we can check if we need to pop out of raw mode
		if (
			term.RESTORE_TERM in rawline
			or term.DISABLE_ALT_SCREEN in rawline
			# Fix for Dialog Debconf Frontend https://gitlab.com/volian/nala/-/issues/211
		) and term.ENABLE_ALT_SCREEN not in rawline:
			self.raw = False
			term.restore_mode()
			self.live.start()
		self.set_last_line(rawline)

	def set_last_line(self, rawline: bytes) -> None:
		"""Set the current line to last line if there is no backspace."""
		# When at the conf prompt if you press Y, then backspace, then hit enter
		# Things get really buggy so instead we check for a backspace
		if term.BACKSPACE not in rawline:
			self.last_line = rawline

	def advance_progress(self) -> None:
		"""Advance the dpkg progress bar."""
		dpkg_progress.advance(self.task)
		if not arguments.scroll:
			self.live.update(
				Panel.fit(
					dpkg_progress.get_renderable(),
					border_style="bold green",
					padding=(0, 0),
				),
				refresh=True,
			)

	def raw_init(self) -> None:
		"""Initialize raw terminal output."""
		if self.raw:
			return
		self.live.raw_init()
		term.set_raw()
		self.raw = True


def check_line_spam(line: str, rawline: bytes, last_line: bytes) -> bool:
	"""Check for, and handle, notices and spam."""
	for message in NOTICES:
		if message in rawline and line not in notice:
			notice.append(line)
			return False
	if b"but it can still be activated by:" in last_line:
		notice.append(f"  {line}")
		return False

	return any(item in line for item in SPAM)


def check_error(data: bytes, line: str, error_in_list: bool = False) -> None:
	"""Check dpkg errors and store them if we need too."""
	# Check so we don't duplicate error messages
	if error_in_list:
		return
	for error in DPKG_ERRORS:
		# Make sure that the error is not spam
		if error in data and all(item not in line for item in SPAM):
			dpkg_error.append(line)


def paren_color(match: Match[str]) -> str:
	"""Color parenthesis."""
	return color("(") if match.group(0) == "(" else color(")")


def line_replace(line: str, header: str) -> str:
	"""Replace wrapper for removing header."""
	return line.replace(header, "").strip()


def format_version(match: list[str], line: str) -> str:
	"""Format version numbers."""
	for ver in match:
		version = ver[1:-1]
		if version and version[0].isdigit():
			new_ver = ver.replace(version, color(version, "BLUE"))
			new_ver = re.sub(PARENTHESIS_PATTERN, paren_color, new_ver)
			line = line.replace(ver, new_ver)
	return line


def fill_pulse(pulse: list[str]) -> str:
	"""Fill the pulse message."""
	last = len(pulse) - 1
	fill = sum(len(line) for line in pulse) + last
	# Set fill width to fit inside the rich panel
	fill = (term.columns - fill) - 5 if arguments.verbose else (term.columns - fill) - 7
	# Minus 2 more to account for ascii simpleDots on the spinner
	if term.console.options.ascii_only:
		fill -= 2
	# Make sure we insert the filler in the right spot
	# In case of extra 1 min as shown below.
	# ['2407', 'kB/s', '30s']
	# ['895', 'kB/s', '1min', '18s']
	index = last - 2
	if "/s" in pulse[index]:
		index = last - 3

	pulse.insert(index, " " * fill)
	return " ".join(pulse)


def msg_formatter(line: str) -> str:
	"""Format dpkg output."""
	if line.endswith("..."):
		line = line.replace("...", "")

	if line.startswith(REMOVING):
		line = REMOVING_MSG.format(
			removing=REMOVING_HEAD, dpkg_msg=line_replace(line, REMOVING)
		)
	elif line.startswith(UNPACKING):
		line = UNPACKING_MSG.format(
			unpacking=UNPACKING_HEAD, dpkg_msg=line_replace(line, UNPACKING)
		)
	elif line.startswith(SETTING_UP):
		line = SETTING_UP_MSG.format(
			setting_up=SETTING_UP_HEAD, dpkg_msg=line_replace(line, SETTING_UP)
		)
	elif line.startswith(PROCESSING):
		line = PROCESSING_MSG.format(
			processing=PROCESSING_HEAD, dpkg_msg=line_replace(line, PROCESSING)
		)
	elif line.startswith(GET):
		line = f"{color(f'{FETCHED}:', 'BLUE')} {' '.join(line.split()[1:])}"

	if match := re.findall(VERSION_PATTERN, line):
		return format_version(match, line)
	return line


class DpkgLive(Live):
	"""Subclass for dpkg live display."""

	def __init__(self, install: bool = True) -> None:
		"""Subclass for dpkg live display."""
		super().__init__(auto_refresh=False, refresh_per_second=4)
		self.install = install
		self.scroll_list: list[str] = []
		self.scroll_config = (False, False, True)
		self.used_scroll: bool = False

	def __enter__(self) -> DpkgLive:
		"""Start the live display."""
		self.start(refresh=self._renderable is not None)
		return self

	def scroll_bar(  # pylint: disable=too-many-arguments
		self,
		msg: str = "",
		apt_fetch: bool = False,
		update_spinner: bool = False,
		use_bar: bool = True,
		rerender: bool = False,
	) -> None:
		"""Print msg to our scroll bar live display."""
		if rerender:
			if not self.used_scroll:
				return
			apt_fetch, update_spinner, use_bar = self.scroll_config
		else:
			self.used_scroll = True
			self.scroll_config = (apt_fetch, update_spinner, use_bar)

		if msg:
			self.scroll_list.append(msg)
		self.slice_list()

		table = Table.grid()
		table.add_column(no_wrap=True, width=term.columns, overflow=OVERFLOW)

		for item in self.scroll_list:
			table.add_row(from_ansi(item))

		if use_bar or update_spinner:
			table.add_row(
				Panel(
					self.get_group(update_spinner, use_bar),
					padding=(0, 0),
					border_style="bold blue" if arguments.scroll else "bold green",
				)
			)

		# We don't need to build the extra panel if we're not scrolling
		if not arguments.scroll:
			self.update(table, refresh=True)
			return

		self.update(
			Panel(
				table,
				title=self.get_title(self.install, apt_fetch),
				title_align="left",
				padding=(0, 0),
				border_style="bold green",
			),
			refresh=True,
		)

	@staticmethod
	def get_title(install: bool, apt_fetch: bool) -> str:
		"""Get the title for our panel."""
		msg = "[bold default]"
		if arguments.command and install and not apt_fetch:
			if arguments.command in ("remove", "purge", "autoremove", "autopurge"):
				if arguments.is_purge():
					msg += _("Purging Packages")
				else:
					msg += _("Removing Packages")
			elif arguments.command in ("upgrade", "dist-upgrade", "full-upgrade"):
				msg += _("Updating Packages")
			elif arguments.command == "install":
				msg += _("Installing Packages")
			elif arguments.command == "history":
				title = (
					_("History Undo")
					if arguments.history == "undo"
					else _("History Redo")
				)
				msg += f"{title} {arguments.history_id}"
			return msg
		if install and apt_fetch:
			return msg + _("Fetching Missed Packages")
		if not arguments.command and arguments.fix_broken:
			return msg + _("Fixing Broken Packages")
		return msg + _("Updating Package List")

	@staticmethod
	def get_group(update_spinner: bool, use_bar: bool) -> RenderableType:
		"""Get the group for our panel."""
		if spinner.text.plain and update_spinner:  # type: ignore[union-attr]
			return (
				Group(spinner, dpkg_progress.get_renderable()) if use_bar else spinner
			)

		return dpkg_progress.get_renderable()

	def slice_list(self) -> None:
		"""Set scroll bar to take up only 1/2 of the screen."""
		scroll_lines = term.lines // 2
		size = len(self.scroll_list)
		if size > scroll_lines and size > 10:
			total = size - max(scroll_lines, 10)
			self.scroll_list = self.scroll_list[total:]

	def raw_init(self) -> None:
		"""Set up the live display to be stopped."""
		# Stop the live display from Auto Refreshing
		if self._refresh_thread:
			self._refresh_thread.stop()
			self._refresh_thread = None

		# We update the live display to blank before stopping it
		self.update("", refresh=True)
		self.stop()

	def start(self, refresh: bool = False) -> None:
		"""Start live rendering display.

		Args:
		----
		        refresh (bool, optional): Also refresh. Defaults to False.

		"""
		with self._lock:
			if self._started:
				return
			self.console.set_live(self)
			self._started = True
			if self._screen:
				self._alt_screen = self.console.set_alt_screen(True)
			self.console.show_cursor(False)
			self._enable_redirect_io()
			self.console.push_render_hook(self)
			if refresh:
				self.refresh()
			if self.auto_refresh:
				self._refresh_thread = Thread(self, self.refresh_per_second)
				self._refresh_thread.start()


def fork() -> tuple[int, int]:
	"""Fork pty or regular."""
	return (os.fork(), 0) if arguments.raw_dpkg else pty.fork()


class AptExpect(fdspawn):  # type: ignore[misc,type-arg,arg-type]
	"""Subclass of fdspawn to add the interact method."""

	def interact(self, install_progress: InstallProgress) -> None:
		"""Hacked up interact method.

		Because pexpect doesn't want to have one for fdspawn.

		This gives control of the child process to the interactive user (the
		human at the keyboard). Keystrokes are sent to the child process, and
		the stdout and stderr output of the child process is printed. This
		simply echos the child stdout and child stderr to the real stdout and
		it echos the real stdin to the child stdin.
		"""
		# Flush the buffer.
		self.write_to_stdout(self.buffer)
		self.stdout.flush()
		self._buffer = self.buffer_type()  # type: ignore[operator]

		_setwinsize(self.child_fd, term.lines, term.columns)

		self.interact_copy(install_progress)

	def interact_copy(self, install_progress: InstallProgress) -> None:
		"""Interact with the pty."""
		while self.isalive():
			try:
				ready = poll_ignore_interrupts([self.child_fd, term.STDIN])
				if self.child_fd in ready and not self._read(install_progress):
					break
				if term.STDIN in ready:
					self._write(install_progress)
			except KeyboardInterrupt:
				term.write(term.CURSER_UP + term.CLEAR_LINE)
				eprint(
					_("{warning} Quitting now could break your system!").format(
						warning=WARNING_PREFIX
					)
				)
				eprint(color(_("Ctrl+C twice quickly will exit") + ELLIPSIS, "RED"))
				sleep(0.5)

	def _read(self, install_progress: InstallProgress) -> bool:
		"""Read data from the pty and send it for formatting."""
		try:
			# We need to read 4096 as sometimes dpkg gives us a lot of lines at one time.
			data = os.read(self.child_fd, 4096)
		except OSError as err:
			if err.args[0] == errno.EIO:
				# Linux-style EOF
				return False
			raise
		if data == b"":
			# BSD-style EOF
			return False
		install_progress.pre_filter(data)
		return True

	def _write(self, install_progress: InstallProgress) -> None:
		"""Write user inputs into the pty."""
		data = os.read(term.STDIN, 1000)
		# Term up and clear in case we answer a question. This stops some live window artifacts.
		# We need to not do this if we're in raw mode or else it breaks things.
		if not install_progress.raw:
			term.write((term.CURSER_UP + term.CLEAR_LINE) * 2)
		while data != b"" and self.isalive():
			split = os.write(self.child_fd, data)
			data = data[split:]



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""The main module for Nala."""
from __future__ import annotations

import errno
import sys

# Imports to get all of the subcommands into typer
import nala.fetch as _fetch  # pylint: disable=unused-import
import nala.nala as _nala  # pylint: disable=unused-import
from nala import _
from nala.constants import ERROR_PREFIX
from nala.options import nala
from nala.utils import eprint


def main() -> None:
	"""Nala function to reference from the entry point."""
	try:
		nala()
	except KeyboardInterrupt:
		eprint("\n" + _("Exiting at your request."))
		sys.exit(130)
	except BrokenPipeError:
		sys.stderr.close()
	except OSError as error:
		if error.errno == errno.ENOSPC:
			sys.exit(_("{error} No space left on device.").format(error=ERROR_PREFIX))
		raise error from error



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""The options module."""
from __future__ import annotations

import os
import sys
from pydoc import pager
from subprocess import run
from typing import Dict, List, NoReturn, Optional, Union, cast

import tomli
import typer
from apt_pkg import config as apt_config

from nala import ROOT, _, __version__, color
from nala.constants import ERROR_PREFIX, GPL3_LICENSE, NOTICE_PREFIX

HookType = Dict[str, Union[str, Dict[str, Union[str, List[str]]]]]


class Config:
	"""Class for managing configurations."""

	def __init__(self) -> None:
		"""Class for managing configurations."""
		self.conf = f"{ROOT}/etc/nala/nala.conf"
		self.data: dict[str, dict[str, str | bool]] = {"Nala": {}}
		self.apt = apt_config

	def read_config(self) -> None:
		"""Read the configuration file."""
		try:
			with open(self.conf, "rb") as file:
				self.data = tomli.load(file)
		except (tomli.TOMLDecodeError, FileNotFoundError) as error:
			print(f"{ERROR_PREFIX} {error}")
			print(
				_(
					"{notice} Unable to read config file: {filename}. Using defaults"
				).format(
					notice=NOTICE_PREFIX,
					filename=color(self.conf, "YELLOW"),
					file=sys.stderr,
				)
			)

	@staticmethod
	def key_error(key: object, value: object) -> NoReturn:
		"""Exit with key error."""
		sys.exit(
			_("{error} Config key '{key}' should be a bool not {value}").format(
				error=ERROR_PREFIX, key=key, value=value
			)
		)

	def get_bool(self, key: str, default: bool = False) -> bool:
		"""Get Boolean from config."""
		value = self.data["Nala"].get(key, default)
		if isinstance(value, bool):
			return value
		self.key_error(key, value)

	def get_str(self, key: str, default: str = "") -> str:
		"""Get String from config."""
		value = self.data["Nala"].get(key, default)
		if isinstance(value, str):
			return value
		self.key_error(key, value)

	def get_hook(self, key: str) -> HookType:
		"""Get Install Hooks from config."""
		return cast(
			HookType,
			self.data.get(key, {}),
		)

	def set(self, key: str, value: str | bool) -> None:
		"""Set value in the Nala Config."""
		self.data["Nala"][key] = value

	# # This will likely need a lot of testing to be able to do something like this
	# # For now it shall remain disabled indefinitely
	# #
	# def recurse_config(self, key: str, conf: dict):
	# 	"""Recurse nala config and pass through apt options.

	# 	Example of toml apt option::

	# 	        APT.Get.AllowUnauthenticated = true
	# 	"""
	# 	for new_key, value in conf.items():
	# 		final_key = new_key
	# 		if key:
	# 			final_key = f"{key}::{new_key}"
	# 		if isinstance(value, dict):
	# 			self.recurse_config(final_key, value)
	# 			continue
	# 		if value == False:
	# 			value = "false"
	# 		elif value == True:
	# 			value = "true"
	# 		self.apt.set(final_key, value)


class Arguments:
	"""Arguments class."""

	# pylint: disable=too-many-instance-attributes, too-many-public-methods
	def __init__(self) -> None:
		"""Arguments class."""
		self.command: str = ""
		self.config = Config()
		self.config.read_config()
		# True Global
		self.verbose: bool
		self.debug: bool

		# Semi Global
		self.download_only: bool
		self.install_recommends: bool
		self.install_suggests: bool
		self.remove_essential: bool
		self.assume_yes: bool
		self.assume_no: bool
		self.update: bool
		self.raw_dpkg: bool
		self.purge: bool = False
		self.fix_broken: bool
		self.simple_summary: bool

		# Used in Show, List and Search
		self.all_versions: bool

		# Used in Search
		self.all_arches: bool

		# Used in Upgrade
		self.full_upgrade: bool

		# Search and List Arguments
		self.names: bool
		self.upgradable: bool
		self.installed: bool
		self.virtual: bool
		self.full: bool

		self.history: str | None
		self.history_id: str

		self.scroll: bool
		self.auto_remove: bool
		self.init_config()

	def __str__(self) -> str:
		"""Return the state of the object as a string."""
		kwarg = "\n    ".join(
			(f"{key} = {value},") for key, value in self.__dict__.items()
		)
		return f"Options = [\n    {kwarg}\n]"

	def set_verbose(self, value: bool) -> None:
		"""Set option."""
		if not value:
			self.verbose = False
			return
		self.verbose = True
		self.scroll = False

	def set_auto_remove(self, value: bool) -> None:
		"""Set option."""
		if value is None:
			return
		self.auto_remove = value

	def set_summary(self, value: bool) -> None:
		"""Set option."""
		if value is None:
			self.simple_summary = self.config.get_bool("simple_summary")
			return
		self.simple_summary = value

	def set_purge(self, value: bool) -> None:
		"""Set option."""
		self.purge = value

	def set_remove_essential(self, value: bool) -> None:
		"""Set option."""
		self.remove_essential = value

	def set_download_only(self, value: bool) -> None:
		"""Set option."""
		self.download_only = value

	def set_fix_broken(self, value: bool) -> None:
		"""Set option."""
		self.fix_broken = value

	def set_full_upgrade(self, value: bool) -> None:
		"""Set option."""
		if value is None:
			self.full_upgrade = self.config.get_bool("full_upgrade")
			return
		self.full_upgrade = value

	def set_color(self, value: bool) -> None:
		"""Set color or no color."""
		if value is None:
			return
		if value:
			os.environ["FORCE_COLOR"] = "true"
		else:
			os.environ["NO_COLOR"] = "true"

	def set_assume_prompt(self, value: Optional[bool]) -> None:
		"""Set option."""
		if value is None:
			self.assume_yes = self.config.get_bool("assume_yes")
			self.assume_no = False
		else:
			self.assume_yes = value
			self.assume_no = not value
			# If the configuration is set to true
			# -y, --assume_yes becomes a toggle to the default behavior
			if self.config.get_bool("assume_yes"):
				self.assume_yes = False

	def set_raw_dpkg(self, value: bool) -> None:
		"""Set option."""
		self.raw_dpkg = value

	def set_all_versions(self, value: bool) -> None:
		"""Set option."""
		self.all_versions = value

	def set_all_arches(self, value: bool) -> None:
		"""Set option."""
		self.all_arches = value

	def set_names(self, value: bool) -> None:
		"""Set option."""
		self.names = value

	def set_installed(self, value: bool) -> None:
		"""Set option."""
		self.installed = value

	def set_upgradable(self, value: bool) -> None:
		"""Set option."""
		if not value and hasattr(self, "upgradable"):
			return
		self.upgradable = value

	def set_virtual(self, value: bool) -> None:
		"""Set option."""
		self.virtual = value

	def set_full(self, value: bool) -> None:
		"""Set option."""
		self.full = value

	def set_recommends(self, value: Optional[bool]) -> None:
		"""Set option."""
		if value is None:
			self.install_recommends = self.config.apt.find_b(
				"APT::Install-Recommends", True
			)
			return
		self.install_recommends = value
		if value:
			self.config.apt.set("APT::Install-Recommends", "1")
		if not value:
			self.config.apt.set("APT::Install-Recommends", "0")

	def set_suggests(self, value: Optional[bool]) -> None:
		"""Set option."""
		if value is None:
			self.install_suggests = self.config.apt.find_b(
				"APT::Install-Suggests", False
			)
			return
		self.install_suggests = value
		if value:
			self.config.apt.set("APT::Install-Suggests", "1")
		else:
			self.config.apt.set("APT::Install-Suggests", "0")

	def set_default_release(self, value: str) -> None:
		"""Set option."""
		if not value:
			return
		self.config.apt.set("APT::Default-Release", value)

	def set_update(self, value: Optional[bool]) -> None:
		"""Set option."""
		if value is None:
			return
		self.update = value

	def set_debug(self, value: bool) -> None:
		"""Set option."""
		self.debug = value

	def set_nala_option(self, key: str, value: str) -> None:
		"""Set Nala option."""
		if value == "false":
			option: str | bool = False
		elif value == "true":
			option = True
		else:
			option = value
		self.config.set(key.split("::", 1)[1], option)

	def set_dpkg_option(self, value: List[str]) -> List[str]:
		"""Set option."""
		if not value:
			return value
		try:
			for opt in value:
				dpkg, option = opt.split("=", 1)
				if dpkg.startswith("Nala::"):
					self.set_nala_option(dpkg, option.strip('"'))
					continue
				self.config.apt.set(dpkg, option.strip('"'))
			# Reinitialize Nala configs in case a Nala option was given
			self.init_config()
		except ValueError:
			sys.exit(
				_("{error} Option {option}: Configuration item must have a '='").format(
					error=ERROR_PREFIX, option=opt
				)
			)
		return value

	def init_config(self) -> None:
		"""Initialize Nala Configs."""
		self.scroll = self.config.get_bool("scrolling_text", True)
		self.auto_remove = self.config.get_bool("auto_remove", True)
		try:
			self.update = (
				self.config.get_bool("auto_update", True)
				if sys.argv[1] in {"upgrade", "full-upgrade", "dist-upgrade"}
				else False
			)
		except IndexError:
			self.update = self.config.get_bool("auto_update", True)

	def state(self) -> str:
		"""Return the state of the object as a string."""
		return f"{self}"

	def is_purge(self) -> bool:
		"""Return if we are to be purging or not."""
		return bool(self.purge or "purge" in self.command)


arguments = Arguments()
nala = typer.Typer(add_completion=True)
history_typer = typer.Typer(name="history")
nala.add_typer(history_typer)


def print_license(value: bool) -> None:
	"""Print the GPLv3 with `--license`."""
	if not value:
		return
	if GPL3_LICENSE.exists():
		with open(GPL3_LICENSE, encoding="utf-8") as file:
			pager(file.read())
	else:
		print(
			_(
				"It seems the system has no license file\n"
				"The full GPLv3 can be found at:\n"
				"https://www.gnu.org/licenses/gpl-3.0.txt"
			)
		)
	sys.exit()


def version(value: bool) -> None:
	"""Print version."""
	if not value:
		return
	print(f"nala {__version__}")
	sys.exit()


def help_callback(value: bool) -> None:
	"""Show man page instead of normal help."""
	if value:
		# This works for purge/remove and autopurge/remove
		command = arguments.command.replace("purge", "remove")
		if command in {"full-upgrade", "dist-upgrade"}:
			command = "upgrade"
		elif command == "uninstall":
			command = "remove"
		elif command == "info":
			command = "show"
		sys.exit(
			run(  # pylint: disable=subprocess-run-check
				["man", f"nala-{command}"]
			).returncode
		)


VERSION = typer.Option(
	False,
	"--version",
	callback=version,
	is_eager=True,
	help=_("Show program's version number and exit."),
)

LICENSE = typer.Option(
	False,
	"--license",
	callback=print_license,
	is_eager=True,
	help=_("Reads the GPLv3 which Nala is licensed under."),
)

VERBOSE = typer.Option(
	False,
	"-v",
	"--verbose",
	callback=arguments.set_verbose,
	is_eager=True,
	help=_("Disable scrolling text and print extra information."),
)

DEBUG = typer.Option(
	False,
	"--debug",
	callback=arguments.set_debug,
	is_eager=True,
	help=_("Logs extra information for debugging."),
)

AUTO_REMOVE = typer.Option(
	None,
	"--autoremove / --no-autoremove",
	callback=arguments.set_auto_remove,
	is_eager=True,
	help=_("Toggle autoremoving packages."),
)

RECOMMENDS = typer.Option(
	None,
	callback=arguments.set_recommends,
	is_eager=True,
	help=_("Toggle installing recommended packages."),
)

SUGGESTS = typer.Option(
	None,
	callback=arguments.set_suggests,
	is_eager=True,
	help=_("Toggle installing suggested packages."),
)

DEFAULT_RELEASE = typer.Option(
	None,
	"-t",
	"--target-release",
	callback=arguments.set_default_release,
	help=_("Set the default release to install packages from."),
)

FULL_UPGRADE = typer.Option(
	None,
	"--full / --no-full",
	callback=arguments.set_full_upgrade,
	help=_("Toggle full-upgrade"),
)

COLOR = typer.Option(
	None,
	"--color / --no-color",
	callback=arguments.set_color,
	help=(_("Force color or no color.")),
)

UPDATE = typer.Option(
	None,
	callback=arguments.set_update,
	help=_("Toggle updating the package list."),
)

SIMPLE = typer.Option(
	None,
	callback=arguments.set_summary,
	is_eager=True,
	help=_("Toggle a more condensed transaction summary."),
)

PURGE = typer.Option(
	False,
	"--purge",
	callback=arguments.set_purge,
	is_eager=True,
	help=_("Purge any packages that would be removed."),
)

CONFIG = typer.Option(
	False,
	"--config",
	help=_("Purge packages not installed that have config files."),
)

REMOVE_ESSENTIAL = typer.Option(
	False,
	"--remove-essential",
	callback=arguments.set_remove_essential,
	is_eager=True,
	help=_("Allow the removal of essential packages."),
)

FIX_BROKEN = typer.Option(
	True,
	"-f",
	"--fix-broken / --no-fix-broken",
	callback=arguments.set_fix_broken,
	is_eager=True,
	help=_("Toggle fix broken packages."),
)

ASSUME_YES = typer.Option(
	None,
	"-y / -n",
	"--assume-yes / --assume-no",
	callback=arguments.set_assume_prompt,
	is_eager=True,
	help=_("Assume 'yes' or 'no' to all prompts."),
)

OPTION = typer.Option(
	[],
	"-o",
	"--option",
	callback=arguments.set_dpkg_option,
	is_eager=True,
	help=_('Set options like Dpkg::Options::="--force-confnew".'),
)

RAW_DPKG = typer.Option(
	False,
	"--raw-dpkg",
	callback=arguments.set_raw_dpkg,
	is_eager=True,
	help=_("Skips all formatting and you get raw dpkg output."),
)

ALL_VERSIONS = typer.Option(
	False,
	"-a",
	"--all-versions",
	callback=arguments.set_all_versions,
	is_eager=True,
	help=_("Show all versions of a package."),
)

ALL_ARCHES = typer.Option(
	False,
	"-A",
	"--all-arches",
	callback=arguments.set_all_arches,
	is_eager=True,
	help=_("Show all architectures of a package."),
)

DOWNLOAD_ONLY = typer.Option(
	False,
	"-d",
	"--download-only",
	callback=arguments.set_download_only,
	is_eager=True,
	help=_("Packages are only retrieved, not unpacked or installed."),
)

NAMES = typer.Option(
	False,
	"-n",
	"--names",
	callback=arguments.set_names,
	is_eager=True,
	help=_("Search only package names."),
)

INSTALLED = typer.Option(
	False,
	"-i",
	"--installed",
	callback=arguments.set_installed,
	is_eager=True,
	help=_("Only installed packages."),
)

NALA_INSTALLED = typer.Option(
	False,
	"-N",
	"--nala-installed",
	is_eager=True,
	help=_("Only packages explicitly installed with Nala."),
)

UPGRADABLE = typer.Option(
	False,
	"-u",
	"--upgradable",
	callback=arguments.set_upgradable,
	is_eager=True,
	help=_("Only upgradable packages."),
)

UPGRADEABLE = typer.Option(
	False,
	"--upgradeable",
	callback=arguments.set_upgradable,
	is_eager=True,
	hidden=True,
)

VIRTUAL = typer.Option(
	False,
	"-V",
	"--virtual",
	callback=arguments.set_virtual,
	is_eager=True,
	help=_("Only virtual packages."),
)

FULL = typer.Option(
	False,
	"--full",
	callback=arguments.set_full,
	is_eager=True,
	help=_("Print the full description of each package."),
)

LISTS = typer.Option(
	False,
	"--lists",
	help=_("Remove package lists located in `/var/lib/apt/lists/`."),
)

FETCH = typer.Option(
	False,
	"--fetch",
	help=_("Remove `nala-sources.list`."),
)

AUTO = typer.Option(
	False, help=_("Run fetch uninteractively. Will still prompt for overwrite")
)

MAN_HELP = typer.Option(
	False,
	"-h",
	"--help",
	callback=help_callback,
	is_eager=True,
	help=_("Show this message and exit."),
)

CONTEXT_SETTINGS = {
	"help_option_names": ["-h", "--help"],
}


@nala.command("help", hidden=True)
@nala.callback(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
# pylint: disable=unused-argument
def global_options(
	ctx: typer.Context,
	_version: bool = VERSION,
	_license: bool = LICENSE,
) -> None:
	"""Each command has its own help page.

	For Example: `nala history --help`
	"""
	if ctx.invoked_subcommand:
		arguments.command = ctx.invoked_subcommand
		if arguments.command == "help":
			print(ctx.get_help())
			sys.exit()
